"""Pytorch Ultrasound Beamforming Library.

Initialize modules for registries.
"""

import torch


def func_on_device(func, device, *args, **kwargs):
    """Moves all tensor arguments of a function to a specified device before calling it.

    Args:
        func (callable): Function to be called.
        device (str or torch.device): Device to move tensors to.
        *args: Positional arguments to be passed to the function.
        **kwargs: Keyword arguments to be passed to the function.
    Returns:
        The output of the function.
    """
    if device is None:
        return func(*args, **kwargs)

    if isinstance(device, str):
        device = torch.device(device)

    def move_to_device(x):
        if isinstance(x, torch.Tensor):
            return x.to(device)
        elif isinstance(x, (list, tuple)):
            return type(x)(move_to_device(i) for i in x)
        elif isinstance(x, dict):
            return {k: move_to_device(v) for k, v in x.items()}
        else:
            return x

    args = move_to_device(args)
    kwargs = move_to_device(kwargs)

    return func(*args, **kwargs)



"""Functionality to convert Verasonics MATLAB workspace to the zea format.

Example of saving the entire workspace to a .mat file (MATLAB):

    .. code-block:: matlab

        >> setup_script;
        >> VSX;
        >> save_raw('C:/path/to/raw_data.mat');

Then convert the saved `raw_data.mat` file to zea format using the following code (Python):

    .. code-block:: python

        from zea.data.convert.verasonics import VerasonicsFile

        VerasonicsFile("C:/path/to/raw_data.mat").to_zea("C:/path/to/output.hdf5")

Or alternatively, use the script below to convert all .mat files in a directory:

    .. code-block:: bash

        python zea/data/convert/verasonics.py "C:/path/to/directory"

or without the directory argument, the script will prompt you to select a directory
using a file dialog.

Event structure
---------------

By default the zea dataformat saves all the data to an hdf5 file with the following structure:

.. code-block:: text

    regular_zea_dataset.hdf5
    ├── data
    └── scan
          └── center_frequency: 1MHz

The data is stored in the ``data`` group and the scan parameters are stored in the ``scan``.
However, when we do an adaptive acquisition, some scanning parameters might change. These
blocks of data with consistent scanning parameters we call events. In the case we have multiple
events, we store the data in the following structure:

.. code-block:: text

    zea_dataset.hdf5
    ├── event_0
    │   ├── data
    │   └── scan
    │       └── center_frequency: 1MHz
    ├── event_1
    │   ├── data
    │   └── scan
    │       └── center_frequency: 2MHz
    ├── event_2
    │   ├── data
    │   └── scan
    └── event_3
        ├── data
        └── scan

This structure is supported by the zea toolbox. The way we can save the data in this structure
from the Verasonics, is by changing the setup script to keep track of the TX struct at each event.

The way this is done is still in development, an example of such an acquisition script that is
compatible with saving event structures is found here:
`setup_agent.m <https://github.com/tue-bmd/needle-tracking/blob/ius2024-demo-nc/verasonics/setup_agent.m>`_

Adding additional elements
--------------------------

You can add additional elements to the dataset by defining a function that reads the
data from the file and returns a ``DatasetElement``. Then pass the function to the
``to_zea`` method as a list.

.. code-block:: python

    def read_max_high_voltage(file):
        lens_correction = file["Trans"]["lensCorrection"][:].item()
        return lens_correction


    def read_high_voltage_func(file):
        return DatasetElement(
            group_name="scan",
            dataset_name="max_high_voltage",
            data=read_max_high_voltage(file),
            description="The maximum high voltage used by the Verasonics system.",
            unit="V",
        )


    VerasonicsFile("C:/path/to/raw_data.mat").to_zea(
        "C:/path/to/output.hdf5",
        [read_high_voltage_func],
    )
"""  # noqa: E501

import os
import sys
import traceback
from pathlib import Path

import h5py
import numpy as np
import yaml
from keras import ops
from schema import And, Optional, Or, Regex, Schema

from zea import log
from zea.data.data_format import DatasetElement, generate_zea_dataset
from zea.func import log_compress, normalize
from zea.internal.device import init_device
from zea.utils import strtobool

_VERASONICS_TO_ZEA_PROBE_NAMES = {
    "L11-4v": "verasonics_l11_4v",
    "L11-5v": "verasonics_l11_5v",
}


_CONVERT_YAML_SCHEMA = Schema(
    {
        "files": [
            {
                "name": str,
                Optional("first_frame"): And(int, lambda x: x >= 0),
                Optional("frames"): Or(
                    "all",
                    And(str, Regex(r"^\d+(-\d+)?$")),  # Matches "30-99" or single number like "5"
                    [And(int, lambda x: x >= 0)],  # List of non-negative integers
                ),
                Optional("transmits"): Or("all", [And(int, lambda x: x >= 0)]),
            }
        ]
    }
)


class VerasonicsFile(h5py.File):
    """HDF5 File class for Verasonics MATLAB workspace files.

    This class extends the h5py.File class to handle Verasonics-specific
    data structures and conventions.
    """

    def dereference_index(self, dataset, index, event=None, subindex=None):
        """Get the element at the given index from the dataset, dereferencing it if
        necessary.

        MATLAB stores items in struct array differently depending on the size. If the size
        is 1, the item is stored as a regular dataset. If the size is larger, the item is
        stored as a dataset of references to the actual data.

        This function dereferences the dataset if it is a reference. Otherwise, it returns
        the dataset.

        Args:
            dataset (h5py.Dataset): The dataset to read the element from.
            index (int): The index of the element to read.
            event (int, optional): The event index. Usually we store each event in the
                second dimension of the dataset. Defaults to None in this case we assume
                that there is only a single event.
            subindex (slice, optional): The subindex of the element to read after
                referencing the actual data. Defaults to None. In this case, all the data
                is returned.
        """
        if isinstance(dataset.fillvalue, h5py.h5r.Reference):
            if event is not None:
                reference = dataset[index, event]
            else:
                reference = dataset[index].item()
            if subindex is None:
                return self[reference][:]
            else:
                return self[reference][subindex]
        else:
            return dataset

    def dereference_all(self, dataset, func=None):
        """Dereference all elements in a dataset.

        Args:
            dataset (h5py.Dataset): The dataset to dereference.
            func (callable, optional): A function to apply to each dereferenced element.

        Returns:
            list: The dereferenced data.
        """
        size = self.get_reference_size(dataset)
        dereferenced_data = []
        for i in range(size):
            element = self.dereference_index(dataset, i)
            element = func(element) if func is not None else element
            dereferenced_data.append(element)
        return dereferenced_data

    @staticmethod
    def get_reference_size(dataset):
        """Get the size of a reference dataset."""
        if isinstance(dataset.fillvalue, h5py.h5r.Reference):
            return len(dataset)
        else:
            return 1

    @staticmethod
    def decode_string(dataset: np.ndarray) -> str:
        """Decode a string dataset."""
        return "".join([chr(c) for c in dataset.squeeze()])

    @staticmethod
    def cast_to_integer(dataset):
        """Cast a h5py dataset to an integer."""
        return int(dataset[:].item())

    @property
    def probe_unit(self):
        """The unit the probe dimensions are defined in."""
        _ALLOWED_UNITS = {"wavelengths", "mm"}
        unit = self.decode_string(self["Trans"]["units"][:])
        assert unit in {"wavelengths", "mm"}, (
            f"Unexpected unit '{unit}' in file, must be one of {_ALLOWED_UNITS}"
        )
        return unit

    @property
    def probe_geometry(self):
        """The probe geometry of shape (n_el, 3)."""
        # Read the probe geometry from the file
        probe_geometry = self["Trans"]["ElementPos"][:3, :]

        # Transpose the probe geometry to have the shape (n_el, 3)
        probe_geometry = probe_geometry.T

        # Convert the probe geometry to meters
        if self.probe_unit == "mm":
            probe_geometry = probe_geometry / 1000
        else:
            probe_geometry = probe_geometry * self.wavelength

        return probe_geometry

    @property
    def wavelength(self):
        """Wavelength of the probe from the file in meters."""

        return self.sound_speed / self.probe_center_frequency

    def read_transmit_events(
        self, event=None, frames="all", allow_accumulate=False, buffer_index=0
    ):
        """Read the events from the file and finds the order in which transmits and receives
        appear in the events.

        Args:
            event (int, optional): The event index. Defaults to None.
            frames (str or list, optional): The frames to read. Defaults to "all".
            allow_accumulate (bool, optional): Sometimes, some transmits are already accumulated
                on the Verasonics system (e.g. harmonic imaging through pulse inversion).
                In this case, the mode in the Receive structure is set to 1 (accumulate).
                If this flag is set to False, an error is raised when such a mode is detected.
            buffer_index (int, optional): The buffer index to read from. Defaults to 0.

        Returns:
            tuple: (tx_order, rcv_order, time_to_next_acq)
                tx_order (list): The order in which the transmits appear in the events.
                rcv_order (list): The order in which the receives appear in the events.
                time_to_next_acq (np.ndarray): The time to next acquisition of shape (n_acq, n_tx).
        """

        num_events = self["Event"]["info"].shape[0]

        # In the Verasonics the transmits may not be in order in the TX structure and a
        # transmit might be reused. Therefore, we need to keep track of the order in which
        # the transmits appear in the Events.
        tx_order = []
        rcv_order = []
        time_to_next_acq = []
        modes = []

        frame_indices = self.get_frame_indices(frames, buffer_index)

        for i in range(num_events):
            # Get the tx
            event_tx = self.dereference_index(self["Event"]["tx"], i)
            event_tx = int(event_tx.item())

            # Get the rcv
            event_rcv = self.dereference_index(self["Event"]["rcv"], i)
            event_rcv = int(event_rcv.item())

            if not bool(event_tx) == bool(event_rcv):
                log.warning(
                    "Events should have both a transmit and a receive or neither. "
                    f"Event {i} has a transmit but no receive or vice versa."
                )

            if not event_tx:
                continue

            # Subtract one to make the indices 0-based
            event_tx -= 1
            event_rcv -= 1

            # Read mode
            mode = self.dereference_index(self["Receive"]["mode"], event_rcv)
            mode = int(mode.item())

            # Check in the Receive structure if this is still the first frame
            framenum = self.dereference_index(self["Receive"]["framenum"], event_rcv)
            framenum = self.cast_to_integer(framenum)

            # Only add the event to the list if it is the first frame since we assume
            # that all frames have the same transmits and receives
            if framenum == 1:
                # Add the event to the list
                tx_order.append(event_tx)
                rcv_order.append(event_rcv)
                modes.append(mode)

            # Read the time_to_next_acq
            seq_control_indices = self.dereference_index(self["Event"]["seqControl"], i)

            for seq_control_index in seq_control_indices:
                seq_control_index = int(seq_control_index.item() - 1)
                seq_control = self.dereference_index(
                    self["SeqControl"]["command"], seq_control_index
                )
                # Decode the seq_control int array into a string
                seq_control = self.decode_string(seq_control)
                if seq_control == "timeToNextAcq":
                    value = self.dereference_index(
                        self["SeqControl"]["argument"], seq_control_index
                    ).item()
                    value = value * 1e-6
                    time_to_next_acq.append(value)

        modes = np.stack(modes)
        tx_order = np.stack(tx_order)
        rcv_order = np.stack(rcv_order)
        time_to_next_acq = np.stack(time_to_next_acq)
        time_to_next_acq = np.reshape(time_to_next_acq, (-1, tx_order.size))

        if np.any(modes == 1) and not allow_accumulate:
            raise ValueError(
                "Some receive events are in accumulate mode (mode=1). "
                "This indicates that the data is already accumulated on the Verasonics system. "
                "Set allow_accumulate=True to allow this."
            )
        elif np.any(modes == 1) and allow_accumulate:
            # We only keep the transmits that are in mode 0 (normal acquisition)
            log.info(
                "Data contains both receives in accumulate mode and replace mode.\n"
                "Discarding transmits in accumulate mode (mode=1). "
                "Keeping transmits in replace mode (mode=0)."
            )
            tx_order = tx_order[modes == 0]
            rcv_order = rcv_order[modes == 0]

            log.info("Dropping time to next acquisition for accumulate mode transmits.")
            time_to_next_acq = None

        if time_to_next_acq is not None:
            if event is not None:
                time_to_next_acq = time_to_next_acq[event]
                time_to_next_acq = np.expand_dims(time_to_next_acq, axis=0)

            time_to_next_acq = time_to_next_acq[frame_indices]

        return tx_order, rcv_order, time_to_next_acq

    def read_t0_delays_apod(self, tx_order, event=None):
        """
        Read the t0 delays and apodization from the file.

        Returns:
            t0_delays (np.ndarray): The t0 delays of shape (n_tx, n_el).
            apod (np.ndarray): The apodization of shape (n_el,).
        """

        t0_delays_list = []
        tx_apodizations_list = []

        for n in tx_order:
            # Get column vector of t0_delays
            if event is None:
                t0_delays = self.dereference_index(self["TX"]["Delay"], n)
            else:
                t0_delays = self.dereference_index(self["TX_Agent"]["Delay"], n, event)
            # Turn into 1d array
            t0_delays = t0_delays[:, 0]

            t0_delays_list.append(t0_delays)

            # Get column vector of apodizations
            if event is None:
                tx_apodizations = self.dereference_index(self["TX"]["Apod"], n)
            else:
                tx_apodizations = self.dereference_index(self["TX_Agent"]["Apod"], n, event)
            # Turn into 1d array
            tx_apodizations = tx_apodizations[:, 0]
            tx_apodizations_list.append(tx_apodizations)

        t0_delays = np.stack(t0_delays_list, axis=0)
        apodizations = np.stack(tx_apodizations_list, axis=0)

        # Convert the t0_delays to seconds
        t0_delays = t0_delays * self.wavelength / self.sound_speed

        return t0_delays, apodizations

    @property
    def sampling_frequency(self):
        """The sampling frequency in Hz from the file."""
        # Read the sampling frequency from the file
        adc_rate = self.dereference_index(self["Receive"]["decimSampleRate"], 0)

        if "quadDecim" in self["Receive"]:
            quaddecim = self.dereference_index(self["Receive"]["quadDecim"], 0)
        else:
            # TODO: Verify if this is correct.
            # On the Vantage NXT the quadDecim field is missing. It seems that it should be
            # set to 1.0 (that decimSampleRate is the actual sampling frequency).
            quaddecim = 1.0

        sampling_frequency = adc_rate / quaddecim * 1e6
        sampling_frequency = sampling_frequency.item()

        if self.is_baseband_mode:
            # Two sequential samples are interpreted as a single complex sample
            # Therefore, we need to halve the sampling frequency
            sampling_frequency = sampling_frequency / 2

        return sampling_frequency

    def read_waveforms(self, tx_order, event=None):
        """
        Read the waveforms from the file.

        Returns:
            waveforms (np.ndarray): The waveforms of shape (n_tx, n_samples).
        """
        waveforms_one_way_list = []
        waveforms_two_way_list = []

        # Read all the waveforms from the file
        n_waveforms = self.get_reference_size(self["TW"]["Wvfm1Wy"])
        for n in range(n_waveforms):
            # Get the row vector of the 1-way waveform
            waveform_one_way = self.dereference_index(self["TW"]["Wvfm1Wy"], n)[:]
            # Turn into 1d array
            waveform_one_way = waveform_one_way[0, :]

            # Get the row vector of the 2-way waveform
            waveform_two_way = self.dereference_index(self["TW"]["Wvfm2Wy"], n)[:]
            # Turn into 1d array
            waveform_two_way = waveform_two_way[0, :]

            waveforms_one_way_list.append(waveform_one_way)
            waveforms_two_way_list.append(waveform_two_way)

        tx_waveform_indices = []

        for n in tx_order:
            # Read the waveform
            if event is None:
                waveform_index = self.dereference_index(self["TX"]["waveform"], n)[:]
            else:
                waveform_index = self.dereference_index(self["TX_Agent"]["waveform"], n, event)[:]
            # Subtract one to make the indices 0-based
            waveform_index -= 1
            # Turn into integer
            waveform_index = int(waveform_index.item())
            tx_waveform_indices.append(waveform_index)

        return tx_waveform_indices, waveforms_one_way_list, waveforms_two_way_list

    def read_beamsteering_angles(self, tx_order, event=None):
        """Beam steering angles in radians (theta, alpha) for each transmit.

        Returns:
            angles (np.ndarray): The beam steering angles of shape (n_tx, 2).
        """
        angles_list = []

        for n in tx_order:
            # Read the polar angle
            if event is None:
                angle = self.dereference_index(self["TX"]["Steer"], n)[:]
            else:
                angle = self.dereference_index(self["TX_Agent"]["Steer"], n, event)[:]

            angles_list.append(angle)
        angles = np.stack(angles_list, axis=0)
        angles = np.squeeze(angles, axis=-1)

        assert angles.shape == (len(tx_order), 2), (
            f"Expected angles shape to be {(len(tx_order), 2)}, but got {angles.shape}"
        )
        return angles

    def read_polar_angles(self, tx_order, event=None):
        """Read the polar angles  of shape (n_tx,) from the file."""
        return self.read_beamsteering_angles(tx_order, event)[:, 0]

    def read_azimuth_angles(self, tx_order, event=None):
        """Read the azimuth angles of shape (n_tx,) from the file."""
        return self.read_beamsteering_angles(tx_order, event)[:, 1]

    @property
    def end_samples(self):
        """The index of the last sample for each receive event."""
        return np.concatenate(self.dereference_all(self["Receive"]["endSample"])).squeeze()

    @property
    def start_samples(self):
        """The index of the first sample for each receive event."""
        return np.concatenate(self.dereference_all(self["Receive"]["startSample"])).squeeze()

    @property
    def n_ax(self):
        """Number of axial samples."""
        n_ax = (self.end_samples - self.start_samples + 1).astype(np.int32)
        n_ax = np.unique(n_ax)
        if n_ax.size != 1:
            raise ValueError(
                "The number of axial samples is not the same for all receive events."
                "We do not support this case yet."
            )
        return n_ax.item()

    @property
    def probe_connector(self):
        """Probe connector indices."""
        probe_connector = self["Trans"]["ConnectorES"][:]
        probe_connector = np.squeeze(probe_connector, axis=0)
        probe_connector = probe_connector.astype(np.int32)
        probe_connector = probe_connector - 1  # make 0-based
        return probe_connector

    @property
    def is_new_save_raw_format(self):
        return "save_raw_version" in self.keys()

    def load_convert_config(self):
        """
        Can load additional conversion configuration from a `convert.yaml` file.

        The `convert.yaml` file should be in the same directory as the .mat file and have
        the following structure:

        .. code-block:: yaml

            files:
            - name: raw_data.mat
              first_frame: 26  # 0-based indexing
              frames: 30-99  # 0-based indexing

        If ``first_frame`` is provided, it will reorder the frames first and use
        the ``frames`` key to subsample afterwards.

        In the example ``frames: 30-99`` means frames 30 to 99 inclusive.

        Returns:
            dict: The configuration for the current file, or an empty dict if no
                configuration is found.
        """
        path = Path(self.filename)
        config_file = path.parent / "convert.yaml"
        if config_file.exists():
            log.info(f"Found convert config file: {log.yellow(config_file)}")
            with open(config_file, "r", encoding="utf-8") as file:
                data = yaml.load(file, Loader=yaml.FullLoader)

            # Validate the YAML structure
            validated_data = _CONVERT_YAML_SCHEMA.validate(data)

            files = validated_data["files"]
            filenames = [file["name"] for file in files]
            if path.name in filenames:
                return files[filenames.index(path.name)]
            elif path.stem in filenames:
                return files[filenames.index(path.stem)]
        return {}

    def get_frame_count(self, buffer_index=0):
        """Get the total number of frames in the RcvBuffer buffer."""
        n_frames = self.dereference_index(self["Resource"]["RcvBuffer"]["numFrames"], buffer_index)
        n_frames = self.cast_to_integer(n_frames)
        return n_frames

    def get_indices_to_reorder(self, first_frame: int, n_frames: int):
        return (np.arange(n_frames) + first_frame) % n_frames

    def get_raw_data_order(self, buffer_index=0):
        """The order of frames in the RcvBuffer buffer.

        Because of the circular buffer used in Verasonics, the frames in the RcvBuffer
        buffer are not necessarily in the correct order. This function computes the
        correct order of frames.
        """
        n_frames = self.get_frame_count(buffer_index)
        try:
            last_frame = self.dereference_index(
                self["Resource"]["RcvBuffer"]["lastFrame"], buffer_index
            )
            last_frame = self.cast_to_integer(last_frame) - 1
            first_frame = (last_frame + 1) % n_frames
        except KeyError:
            log.warning(
                "Could not find 'lastFrame' in 'Resource/RcvBuffer'. "
                "Assuming data is already in correct order."
            )
            return np.arange(n_frames)

        return self.get_indices_to_reorder(first_frame, n_frames)

    def read_raw_data(self, event=None, frames="all", buffer_index=0, first_frame_idx=None):
        """
        Read the raw data from the file.

        Returns:
            raw_data (np.ndarray): The raw data of shape (n_rcv, n_samples).
        """

        # Read the raw data from the file
        if event is None:
            raw_data = self.dereference_index(self["RcvData"], buffer_index)
        else:
            # for now we only index frames as events
            raw_data = self.dereference_index(self["RcvData"], buffer_index, subindex=event)
            raw_data = np.expand_dims(raw_data, axis=0)

        # Convert the raw data to a numpy array to allow out-of-order indexing later
        raw_data = np.asarray(raw_data, dtype=np.int16)

        # Reorder and select channels based on probe elements
        if self.is_new_save_raw_format:
            raw_data = raw_data[:, self.probe_connector, :]
        else:
            log.warning(
                "Data was not saved using the updated `save_raw` function (version >= 1.0). "
                "In that case, we assume that the channel order in the data matches the "
                "probe element order. Please verify that this is correct!"
            )

        # Re-order frames such that sequence is correct
        if first_frame_idx is not None:
            n_frames = self.get_frame_count(buffer_index)
            indices = self.get_indices_to_reorder(first_frame_idx, n_frames)
        else:
            indices = self.get_raw_data_order(buffer_index)
        raw_data = raw_data[indices]

        # Select only the requested frames
        frame_indices = self.get_frame_indices(frames, buffer_index)
        raw_data = raw_data[frame_indices]

        # Trim the raw data to the final sample in the buffer
        final_sample_in_buffer = int(self.end_samples.max())
        raw_data = raw_data[:, :, :final_sample_in_buffer]

        # Determine n_tx based on the final sample in buffer and n_ax
        # For some sequences, transmits are already aggregated in the raw data
        # (e.g. harmonic imaging through pulse inversion)
        n_tx = final_sample_in_buffer // self.n_ax

        # Reshape the raw data to (n_frames, n_el, n_tx, n_ax)
        raw_data = raw_data.reshape((raw_data.shape[0], raw_data.shape[1], n_tx, self.n_ax))

        # Transpose the raw data to (n_frames, n_tx, n_ax, n_el)
        raw_data = np.transpose(raw_data, (0, 2, 3, 1))

        # Add channel dimension
        raw_data = raw_data[..., None]

        # If the data is captured in BS100BW mode or BS50BW mode, the data is stored in
        # as complex IQ data.
        if self.is_baseband_mode:
            raw_data = np.concatenate(
                (
                    raw_data[:, :, 0::2, :, :],
                    -raw_data[:, :, 1::2, :, :],
                ),
                axis=-1,
            )

        return raw_data

    @property
    def probe_center_frequency(self):
        """Center frequency of the probe from the file in Hz."""

        return self["Trans"]["frequency"][:].item() * 1e6

    def read_center_frequency(self, waveform_index):
        """Center frequency of the transmit from the file in Hz."""

        tw_type = self.dereference_index(self["TW"]["type"], waveform_index)[:]
        tw_type = self.decode_string(tw_type)
        if tw_type == "parametric":
            center_freq, _, _, _ = self.dereference_index(
                self["TW"]["Parameters"],
                waveform_index,
            )[:].squeeze()
        else:
            raise ValueError(
                f"Unsupported waveform type '{tw_type}' for center frequency extraction."
            )

        return center_freq.item() * 1e6  # Convert MHz to Hz

    def read_center_frequencies(self, tx_waveform_indices):
        """Center frequencies of the transmits from the file in Hz."""

        center_frequencies = []
        for waveform_index in tx_waveform_indices:
            center_frequency = self.read_center_frequency(waveform_index)
            center_frequencies.append(center_frequency)

        center_frequencies = np.stack(center_frequencies)
        center_frequencies = np.unique(center_frequencies)
        if center_frequencies.size != 1:
            raise ValueError(
                "Multiple center frequencies found in file: "
                f"{center_frequencies}. We do not support this case at the moment."
            )
        return center_frequencies.item()

    @property
    def demodulation_frequency(self):
        """Demodulation frequency of the probe from the file in Hz."""

        demod_freq = self.dereference_all(self["Receive"]["demodFrequency"])
        demod_freq = np.unique(demod_freq)
        assert demod_freq.size == 1, (
            f"Multiple demodulation frequencies found in file: {demod_freq}. "
            "We do not support this case."
        )

        return demod_freq.item() * 1e6

    @property
    def sound_speed(self):
        """Speed of sound in the medium in m/s."""

        return self["Resource"]["Parameters"]["speedOfSound"][:].item()

    def read_initial_times(self, rcv_order):
        """Reads the initial times from the file.

        Args:
            rcv_order (list): The order in which the receives appear in the events.
            wavelength (float): The wavelength of the probe.

        Returns:
            initial_times (np.ndarray): The initial times of shape (n_rcv,).
        """
        initial_times = []
        for n in rcv_order:
            start_depth = self.dereference_index(self["Receive"]["startDepth"], n).item()

            initial_times.append(2 * start_depth * self.wavelength / self.sound_speed)

        return np.stack(initial_times).astype(np.float32)

    @property
    def probe_name(self):
        """The name of the probe from the file."""
        probe_name = self["Trans"]["name"][:]
        probe_name = self.decode_string(probe_name)
        # Translates between verasonics probe names and zea probe names
        if probe_name in _VERASONICS_TO_ZEA_PROBE_NAMES:
            probe_name = _VERASONICS_TO_ZEA_PROBE_NAMES[probe_name]
        else:
            log.warning(
                f"Probe name '{probe_name}' is not in the list of known probes. "
                "Please add it to the _VERASONICS_TO_ZEA_PROBE_NAMES dictionary. "
                "Falling back to generic probe."
            )
            probe_name = "generic"

        return probe_name

    def read_focus_distances(self, tx_order, event=None):
        """Reads the focus distances from the file.

        Args:
            tx_order (list): The order in which the transmits appear in the events.

        Returns:
            focus_distances (list): The focus distances of shape (n_tx,) in meters.
        """
        focus_distances = []
        for n in tx_order:
            if event is None:
                focus_distance = self.dereference_index(self["TX"]["focus"], n)[:].item()
            else:
                focus_distance = self.dereference_index(
                    self["TX_Agent"]["focus"],
                    n,
                    event,
                )[:].item()
            focus_distances.append(focus_distance)

        # Convert focus distances from wavelengths to meters
        focus_distances = np.stack(focus_distances) * self.wavelength

        return focus_distances

    def read_transmit_origins(self, tx_order, event=None):
        """Reads the transmit origins from the file.

        Args:
            tx_order (list): The order in which the transmits appear in the events.

        Returns:
            origins (np.ndarray): The transmit origins of shape (n_tx, 3) in meters.
        """
        origins = []
        for n in tx_order:
            if event is None:
                origin = self.dereference_index(self["TX"]["Origin"], n)
            else:
                origin = self.dereference_index(self["TX_Agent"]["Origin"], n, event)
            origins.append(origin.squeeze())

        # Convert origins from wavelengths to meters
        origins = np.stack(origins) * self.wavelength

        return origins

    @property
    def _probe_geometry_is_ordered_ula(self):
        """Checks if the probe geometry is ordered as a uniform linear array (ULA)."""
        diff_vec = self.probe_geometry[1:] - self.probe_geometry[:-1]
        return np.isclose(diff_vec, diff_vec[0]).all()

    def planewave_focal_distance_to_inf(self, focus_distances, t0_delays, tx_apodizations):
        """Detects plane wave transmits and sets the focus distance to infinity.

        Args:
            focus_distances (np.ndarray): The focus distances of shape (n_tx,).
            t0_delays (np.ndarray): The t0 delays of shape (n_tx, n_el).
            tx_apodizations (np.ndarray): The apodization of shape (n_tx, n_el).

        Returns:
            focus_distances (np.ndarray): The focus distances of shape (n_tx,).

        Note:
            This function assumes that the probe_geometry is a 1d uniform linear array.
            If not it will warn and return.
        """
        if not self._probe_geometry_is_ordered_ula:
            log.warning(
                "The probe geometry is not ordered as a uniform linear array. "
                "Focal distances are not set to infinity for plane waves."
            )
            return focus_distances

        for tx in range(focus_distances.size):
            mask_active = np.abs(tx_apodizations[tx]) > 0
            if np.sum(mask_active) < 2:
                continue
            t0_delays_active = t0_delays[tx][mask_active]

            # If the t0_delays all have the same offset, we assume it is a plane wave
            if np.std(np.diff(t0_delays_active)) < 1e-16:
                focus_distances[tx] = np.inf

        return focus_distances

    @property
    def bandwidth_percent(self):
        """Receive bandwidth as a percentage of center frequency."""
        SUPPORTED_SAMPLE_MODES = ["NS200BW", "BS100BW", "BS67BW", "BS50BW"]

        # For all unique sample modes
        bandwidth_percent = self.dereference_all(
            self["Receive"]["sampleMode"], func=self.decode_string
        )
        bandwidth_percent = set(bandwidth_percent)

        # Ensure only a single bandwidth mode is used
        assert len(bandwidth_percent) == 1, (
            f"Multiple bandwidth modes found in file: {bandwidth_percent}. "
            "We do not support this case."
        )
        bandwidth_percent = bandwidth_percent.pop()

        # Check if the bandwidth mode is supported, and extract the percentage
        assert bandwidth_percent in SUPPORTED_SAMPLE_MODES, (
            f"Unexpected bandwidth mode '{bandwidth_percent}' in file."
            f"Expected one of {SUPPORTED_SAMPLE_MODES}"
        )
        return int(bandwidth_percent[2:-2])

    @property
    def is_baseband_mode(self):
        """If the data is captured in 'BS100BW' mode or 'BS50BW' mode.

        - The data is stored as complex IQ data.
        - The sampling frequency is halved.
        - Two sequential samples are interpreted as a single complex sample.
          Therefore, we need to halve the sampling frequency.
        """
        return self.bandwidth_percent in (50, 100)

    @property
    def lens_correction(self):
        """The lens correction: 1 way delay in wavelengths thru lens"""

        return self["Trans"]["lensCorrection"][:].item()

    @property
    def tgc_gain_curve(self):
        """The TGC gain curve from the file interpolated to the number of axial samples (n_ax,)."""

        gain_curve = self["TGC"]["Waveform"][:][:, 0]

        # Normalize the gain_curve to [0, 40]dB
        gain_curve = gain_curve / 1023 * 40

        # The gain curve is sampled at 800ns (See Verasonics documentation for details.
        # Specifically the tutorial sequence programming)
        gain_curve_sampling_period = 800e-9

        # Define the time axis for the gain curve
        t_gain_curve = np.arange(gain_curve.size) * gain_curve_sampling_period

        # For baseband mode two consecutive samples are combined into a single complex sample
        n_ax = self.n_ax if not self.is_baseband_mode else self.n_ax // 2

        # Define the time axis for the axial samples
        t_samples = np.arange(n_ax) / self.sampling_frequency

        # Interpolate the gain_curve to the number of axial samples
        gain_curve = np.interp(t_samples, t_gain_curve, gain_curve)

        # The gain_curve gains are in dB, so we need to convert them to linear scale
        gain_curve = 10 ** (gain_curve / 20)

        return gain_curve

    def get_image_data_p_frame_order(self, buffer_index=0):
        """The order of frames in the ImgDataP buffer.

        Because of the circular buffer used in Verasonics, the frames in the ImgDataP
        buffer are not necessarily in the correct order. This function computes the
        correct order of frames.
        """
        n_frames = self.dereference_index(
            self["Resource"]["ImageBuffer"]["numFrames"], buffer_index
        )
        n_frames = self.cast_to_integer(n_frames)
        try:
            first_frame = self.dereference_index(
                self["Resource"]["ImageBuffer"]["firstFrame"], buffer_index
            )
            last_frame = self.dereference_index(
                self["Resource"]["ImageBuffer"]["lastFrame"], buffer_index
            )
            first_frame = self.cast_to_integer(first_frame) - 1  # make 0-based
            last_frame = self.cast_to_integer(last_frame) - 1  # make 0-based
            indices = np.arange(first_frame, first_frame + n_frames) % n_frames
            assert indices[-1] == last_frame, (
                "The last frame index does not match the expected last frame index."
            )
            return indices
        except KeyError:
            log.warning(
                "Could not find 'firstFrame' or 'lastFrame' in 'Resource/ImageBuffer'. "
                "Assuming data is already in correct order."
            )
            return np.arange(n_frames)

    def read_image_data_p(self, event=None, frames="all", buffer_index=0):
        """Reads the image data from the file.

        Uses the ``ImgDataP`` buffer, which is used for spatial filtering
        and persistence processing. Generally, this buffer does not contain the same frames
        as the raw data buffer. This happens because the Verasonics often does not reconstruct
        every acquired frame. This means that the images in this buffer often skip frames, and
        span a longer time period than the raw data buffer.

        Returns:
            `image_data` (`np.ndarray`): The image data.
        """
        # Check if the file contains image data
        if "ImgDataP" not in self:
            return None

        # Get the dataset reference
        image_data_ref = self["ImgDataP"][:].squeeze()[buffer_index]
        # Dereference the dataset
        if event is None:
            image_data = self[image_data_ref][:]
        else:
            image_data = self[image_data_ref][event]
            image_data = np.expand_dims(image_data, axis=0)

        # Re-order images such that sequence is correct
        indices = self.get_image_data_p_frame_order(buffer_index)
        image_data = image_data[indices, :, :]

        # Normalize and log-compress the image data
        image_data = normalize(image_data, output_range=(0, 1), input_range=(0, None))
        image_data = log_compress(image_data)
        image_data = ops.convert_to_numpy(image_data)

        # Select only the requested frames
        frame_indices = self.get_frame_indices(frames, buffer_index)
        image_data = image_data[frame_indices]

        return image_data

    @property
    def element_width(self):
        """The element width in meters from the file."""
        element_width = self["Trans"]["elementWidth"][:].item()

        # Convert the probe element width to meters
        if self.probe_unit == "mm":
            element_width = element_width / 1000
        else:
            element_width = element_width * self.wavelength

        return element_width

    def read_verasonics_file(
        self,
        event=None,
        additional_functions=None,
        frames=None,
        allow_accumulate=False,
        buffer_index=0,
    ):
        """Reads data from a .mat Verasonics output file.

        Args:
            event (int, optional): The event index. Defaults to None in this case we assume
                the data file is stored without event structure.
            additional_functions (list, optional): A list of functions that read additional
                data from the file. Each function should take the file as input and return a
                `DatasetElement`. Defaults to None.
            frames (str or list of int, optional): The frames to add to the file. This can be
                a list of integers, a range of integers (e.g. 4-8), or 'all'. Defaults to
                None, which means all frames, unless specified in a `convert.yaml` file.
            allow_accumulate (bool, optional): Sometimes, some transmits are already accumulated
                on the Verasonics system (e.g. harmonic imaging through pulse inversion).
                In this case, the mode in the Receive structure is set to 1 (accumulate).
                If this flag is set to False, an error is raised when such a mode is detected.
            buffer_index (int, optional): The buffer index to read from. Defaults to 0.
        """

        if additional_functions is None:
            additional_functions = []

        convert_config = self.load_convert_config()

        if frames is None:
            frames = convert_config.get("frames", "all")

        first_frame_idx = convert_config.get("first_frame", None)

        tx_order, rcv_order, time_to_next_transmit = self.read_transmit_events(
            frames=frames, allow_accumulate=allow_accumulate, buffer_index=buffer_index
        )
        initial_times = self.read_initial_times(rcv_order)

        # these are capable of handling multiple events
        raw_data = self.read_raw_data(event, frames=frames, first_frame_idx=first_frame_idx)

        polar_angles = self.read_polar_angles(tx_order, event)
        azimuth_angles = self.read_azimuth_angles(tx_order, event)
        t0_delays, tx_apodizations = self.read_t0_delays_apod(tx_order, event)
        focus_distances = self.read_focus_distances(tx_order, event)
        transmit_origins = self.read_transmit_origins(tx_order, event)

        tx_waveform_indices, waveforms_one_way_list, waveforms_two_way_list = self.read_waveforms(
            tx_order, event
        )
        center_frequency = self.read_center_frequencies(tx_waveform_indices)
        focus_distances = self.planewave_focal_distance_to_inf(
            focus_distances, t0_delays, tx_apodizations
        )

        if event is None:
            group_name = "scan"
        else:
            group_name = f"event_{event}/scan"

        additional_elements = []

        # Add Verasonics lens correction to additional elements
        el_lens_correction = DatasetElement(
            group_name=group_name,
            dataset_name="lens_correction",
            data=self.lens_correction,
            description=(
                "The lens correction value used by Verasonics. This value is the "
                "additional path length in wavelength that the lens introduces. "
                "(This disregards refraction.)"
            ),
            unit="wavelengths",
        )
        additional_elements.append(el_lens_correction)

        # Add Verasonics ImgDataP buffer to additional elements
        try:
            verasonics_image_buffer = self.read_image_data_p(event, frames=frames)
            verasonics_image_buffer = DatasetElement(
                dataset_name="verasonics_image_buffer",
                data=verasonics_image_buffer,
                description=(
                    "The Verasonics ImgDataP buffer. "
                    "WARNING: This buffer may skip frames compared to the raw data! "
                    "Use only for reference."
                ),
                unit="unitless",
            )
            additional_elements.append(verasonics_image_buffer)
        except Exception as e:
            log.error(f"Could not read Verasonics ImgDataP buffer: {e}, skipping.")

        # Add additional elements from user-defined functions
        for additional_function in additional_functions:
            additional_elements.append(additional_function(self))

        data = {
            "probe_geometry": self.probe_geometry,
            "time_to_next_transmit": time_to_next_transmit,
            "t0_delays": t0_delays,
            "tx_apodizations": tx_apodizations,
            "sampling_frequency": self.sampling_frequency,
            "polar_angles": polar_angles,
            "azimuth_angles": azimuth_angles,
            "bandwidth_percent": self.bandwidth_percent,
            "raw_data": raw_data,
            "center_frequency": center_frequency,
            "demodulation_frequency": self.demodulation_frequency,
            "sound_speed": self.sound_speed,
            "initial_times": initial_times,
            "probe_name": self.probe_name,
            "focus_distances": focus_distances,
            "transmit_origins": transmit_origins,
            "tx_waveform_indices": tx_waveform_indices,
            "waveforms_one_way": waveforms_one_way_list,
            "waveforms_two_way": waveforms_two_way_list,
            "tgc_gain_curve": self.tgc_gain_curve,
            "element_width": self.element_width,
            "additional_elements": additional_elements,
        }

        return data

    def _parse_frames_argument(self, frames, n_frames):
        value_error = ValueError(
            f"Invalid frames argument: {frames}. "
            "Expected 'all', a range (e.g. '4-8'), or a list of integers."
        )

        if isinstance(frames, str):
            if frames == "all":
                return list(range(n_frames))
            elif "-" in frames:
                start, end = frames.split("-")
                return list(range(int(start), int(end) + 1))
            else:
                # Try to convert to integer
                try:
                    frame_index = int(frames)
                    return [frame_index]
                except ValueError:
                    raise value_error
        elif isinstance(frames, (list, tuple)):
            # Recursively parse each element
            frame_indices = []
            for frame in frames:
                frame_indices.extend(self._parse_frames_argument(frame, n_frames))
            return frame_indices
        elif isinstance(frames, int):
            return [frames]
        else:
            raise value_error

    def get_frame_indices(self, frames, buffer_index=0):
        """Creates a numpy array of frame indices from the file and the frames argument.

        Args:
            frames (str): The frames argument. This can be "all", a range of integers
                (e.g. "4-8"), or a list of frame indices.

        Returns:
            frame_indices (np.ndarray): The frame indices.
        """
        # Read the number of frames from the file
        n_frames = self.get_frame_count(buffer_index)

        frame_indices = self._parse_frames_argument(frames, n_frames)
        frame_indices = np.asarray(frame_indices)
        frame_indices = np.unique(frame_indices)  # Remove duplicates
        frame_indices.sort()  # Sort the indices

        if np.any(frame_indices >= n_frames):
            log.error(
                f"Frame indices {frame_indices} are out of bounds. "
                f"The file contains {n_frames} frames. "
                f"Using only the indices that are within bounds."
            )
            # Remove out of bounds indices
            frame_indices = frame_indices[frame_indices < n_frames]

        return frame_indices

    def to_zea(
        self,
        output_path,
        additional_functions=None,
        frames=None,
        allow_accumulate=False,
        enable_compression=True,
    ):
        """Converts the Verasonics file to the zea format.

        Args:
            output_path (str): The path to the output file (.hdf5 file).
            additional_functions (list, optional): A list of functions that read additional
                data from the file. Each function should take the file as input and return a
                `DatasetElement`. Defaults to None.
            frames (str or list of int, optional): The frames to add to the file. This can be
                a list of integers, a range of integers (e.g. 4-8), or 'all'. Defaults to
                None, which means all frames are used, unless specified otherwise in a
                `convert.yaml` file.
            allow_accumulate (bool, optional): Sometimes, some transmits are already accumulated
                on the Verasonics system (e.g. harmonic imaging through pulse inversion).
                In this case, the mode in the Receive structure is set to 1 (accumulate).
                If this flag is set to False, an error is raised when such a mode is detected.
                Defaults to False.
            enable_compression (bool, optional): Whether to enable compression when saving
                the zea file. Defaults to True.
        """
        if "TX_Agent" in self:
            active_keys = self["TX_Agent"].keys()
            log.info(
                f"Found active imaging data with {len(active_keys)} events. "
                "Will convert and save all parameters for each event separately."
            )
            num_events = set(self["TX_Agent"][key].shape[-1] for key in active_keys)
            assert len(num_events) == 1, (
                "All TX_Agent entries should have the same number of events."
            )
            num_events = num_events.pop()

            # loop over TX_Agent entries and overwrite TX each time
            data = {}
            for event in range(num_events):
                data[event] = self.read_verasonics_file(
                    event=event,
                    additional_functions=additional_functions,
                    allow_accumulate=allow_accumulate,
                )

            # convert dict of events to dict of lists
            data = {key: [data[event][key] for event in data] for key in data[0]}
            description = ["Verasonics data with multiple events"] * num_events

            # Generate the zea dataset
            generate_zea_dataset(
                path=output_path,
                **data,
                event_structure=True,
                description=description,
                enable_compression=enable_compression,
            )

        else:
            # Here we call al the functions to read the data from the file
            log.info("Reading Verasonics file...")
            data = self.read_verasonics_file(
                additional_functions=additional_functions,
                frames=frames,
                allow_accumulate=allow_accumulate,
            )

            # Generate the zea dataset
            log.info("Generating zea dataset...")
            generate_zea_dataset(
                path=output_path,
                **data,
                description="Verasonics data",
                enable_compression=enable_compression,
            )


def _zea_from_verasonics_workspace(input_path, output_path, **kwargs):
    """Helper function around ``VerasonicsFile.to_zea``"""

    # Create the output directory if it does not exist
    input_path = Path(input_path)
    output_path = Path(output_path)

    output_path.parent.mkdir(parents=True, exist_ok=True)

    assert input_path.is_file(), log.error(f"Input file {log.yellow(input_path)} does not exist.")

    # Load the data
    with VerasonicsFile(input_path, "r") as file:
        file.to_zea(output_path, **kwargs)

    log.success(f"Converted {log.yellow(input_path)} to {log.yellow(output_path)}")


def get_answer(prompt, additional_options=None):
    """Get a yes or no answer from the user. There is also the option to provide
    additional options. In case yes or no is selected, the function returns a boolean.
    In case an additional option is selected, the function returns the selected option
    as a string.

    Args:
        prompt (str): The prompt to show the user.
        additional_options (list, optional): Additional options to show the user.
            Defaults to None.

    Returns:
        str: The user's answer.
    """
    while True:
        answer = input(prompt)
        try:
            bool_answer = strtobool(answer)
            return bool_answer
        except ValueError:
            if additional_options is not None and answer in additional_options:
                return answer
        log.warning("Invalid input.")


def convert_verasonics(args):
    """
    Converts a Verasonics MATLAB workspace file (.mat) or a directory containing multiple
    such files to the zea format.

    Args:
        args (argparse.Namespace): An object with attributes:

            - src (str): Source folder path.
            - dst (str): Destination folder path.
            - frames (list[str]): MATLAB frames spec (e.g., ["all"], integers, or ranges like "4-8")
            - allow_accumulate (bool): Whether to allow accumulate mode.
            - device (str): Device to use for processing.
    """

    init_device(args.device)

    # Variable to indicate what to do with existing files.
    # Is set by the user in case these are found.
    existing_file_policy = None

    if args.src is None:
        log.info("Select a directory containing Verasonics MATLAB workspace files.")
        # Create a Tkinter root window
        try:
            import tkinter as tk
            from tkinter import filedialog

            root = tk.Tk()
            root.withdraw()
            # Prompt the user to select a file or directory
            selected_path = filedialog.askdirectory()
        except ImportError as e:
            raise ImportError(
                log.error(
                    "tkinter is not installed. Please install it with 'apt install python3-tk'."
                )
            ) from e
        except Exception as e:
            raise ValueError(
                log.error(
                    "Failed to open a file dialog (possibly in headless state). "
                    "Please provide a path as an argument. "
                )
            ) from e
    else:
        selected_path = args.src

    # Exit when no path is selected
    if not selected_path:
        log.error("No path selected.")
        sys.exit()
    else:
        selected_path = Path(selected_path)

    selected_path_is_directory = os.path.isdir(selected_path)

    # Set the output path to be next to the input directory with _zea appended
    # to the name
    if args.dst is None:
        if selected_path_is_directory:
            output_path = selected_path.parent / (Path(selected_path).name + "_zea")
        else:
            output_path = str(selected_path.with_suffix("")) + "_zea.hdf5"
            output_path = Path(output_path)
    else:
        output_path = Path(args.dst)
        if selected_path.is_file() and output_path.suffix not in (".hdf5", ".h5"):
            log.error(
                "When converting a single file, the output path should have the .hdf5 "
                "or .h5 extension."
            )
            sys.exit()
        elif selected_path.is_dir() and output_path.is_file():
            log.error("When converting a directory, the output path should be a directory.")
            sys.exit()

        if output_path.is_dir() and not selected_path_is_directory:
            output_path = output_path / (selected_path.name + "_zea.hdf5")

    log.info(f"Selected path: {log.yellow(selected_path)}")

    # Do the conversion of a single file
    if not selected_path_is_directory:
        if output_path.is_file():
            answer = get_answer(
                f"File {log.yellow(output_path)} exists. Overwrite?"
                "\n\ty\t - Overwrite"
                "\n\tn\t - Skip"
                "\nAnswer: "
            )
            if answer is True:
                log.warning(f"{selected_path} exists. Deleting...")
                output_path.unlink(missing_ok=False)
            else:
                log.info("Aborting...")
                sys.exit()
        _zea_from_verasonics_workspace(
            selected_path,
            output_path,
            frames=args.frames,
            allow_accumulate=args.allow_accumulate,
            enable_compression=not args.no_compression,
        )
    else:
        # Continue with the rest of your code...
        for root, dirs, files in os.walk(selected_path):
            for mat_file in files:
                # Skip non-mat files
                if not mat_file.endswith(".mat"):
                    continue

                log.info(f"Found raw data file {log.yellow(mat_file)}")

                # Convert the file to a Path object
                mat_file = Path(mat_file)

                # Construct the output path
                relative_path = (Path(root) / Path(mat_file)).relative_to(selected_path)
                file_output_path = output_path / (relative_path.with_suffix(".hdf5"))

                full_path = selected_path / relative_path

                # Handle existing files
                if file_output_path.is_file():
                    if existing_file_policy is None:
                        answer = get_answer(
                            f"File {log.yellow(file_output_path)} exists. Overwrite?"
                            "\n\ty\t - Overwrite"
                            "\n\tn\t - Skip"
                            "\n\tya\t - Overwrite all existing files"
                            "\n\tna\t - Skip all existing files"
                            "\nAnswer: ",
                            additional_options=("ya", "na"),
                        )
                        if answer == "ya":
                            existing_file_policy = "overwrite"
                        elif answer == "na":
                            existing_file_policy = "skip"
                            continue

                    if existing_file_policy == "skip" or answer is False:
                        log.info("Skipping...")
                        continue

                    if existing_file_policy == "overwrite" or answer is True:
                        log.warning(f"{log.yellow(file_output_path)} exists. Deleting...")
                        file_output_path.unlink(missing_ok=False)

                try:
                    _zea_from_verasonics_workspace(
                        full_path,
                        file_output_path,
                        frames=args.frames,
                        allow_accumulate=args.allow_accumulate,
                        enable_compression=not args.no_compression,
                    )
                except Exception:
                    # Print error message without raising it
                    log.error(f"Failed to convert {mat_file}")
                    # Print stacktrace
                    traceback.print_exc()

                    continue



"""Structure containing parameters defining an ultrasound scan.

This module provides the :class:`Scan` class, a flexible structure
for managing all parameters related to an ultrasound scan acquisition.

Features
^^^^^^^^

- **Flexible initialization:** The :class:`Scan` class supports lazy initialization,
  allowing you to specify any combination of supported parameters. You can pass only
  the parameters you have, and the rest will be computed or set to defaults as needed.

- **Automatic computation:** Many scan properties (such as
  grid, number of pixels, wavelength, etc.) are computed automatically from the
  provided parameters. This enables you to work with minimal input and still obtain
  all necessary scan configuration details.

- **Dependency tracking and lazy evaluation:** Derived properties are computed only
  when accessed, and are automatically invalidated and recomputed if their dependencies
  change. This ensures efficient memory usage and avoids unnecessary computations.

- **Parameter validation:** All parameters are type-checked and validated against
  a predefined schema, reducing errors and improving robustness.

- **Selection of transmits:** The scan supports flexible selection of transmit events,
  using the :meth:`set_transmits` method. You can select all, a specific number,
  or specific transmit indices. The selection is stored and can be accessed via
  the :attr:`selected_transmits` property.

Comparison to ``zea.Config`` and ``zea.Probe``
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

- :class:`zea.config.Config`: A general-purpose parameter dictionary for experiment and pipeline
  configuration. It is not specific to ultrasound acquisition and does not compute
  derived parameters.

- :class:`zea.probes.Probe`: Contains only probe-specific parameters (e.g., geometry, frequency).

- :class:`zea.scan.Scan`: Combines all parameters relevant to an ultrasound acquisition,
  including probe, acquisition, and scan region. It also provides automatic computation
  of derived properties and dependency management.

Example Usage
^^^^^^^^^^^^^

.. doctest::

    >>> from zea import Config, Probe, Scan

    >>> # Initialize Scan from a Probe's parameters
    >>> probe = Probe.from_name("verasonics_l11_4v")
    >>> scan = Scan(**probe.get_parameters(), grid_size_z=256, n_tx=11)

    >>> # Or initialize from a Config object
    >>> config = Config.from_path("hf://zeahub/configs/config_picmus_rf.yaml")
    >>> scan = Scan(n_tx=11, **config.scan)

    >>> # Or manually specify parameters
    >>> scan = Scan(
    ...     grid_size_x=128,
    ...     grid_size_z=256,
    ...     xlims=(-0.02, 0.02),
    ...     zlims=(0.0, 0.06),
    ...     ylims=(0.0, 0.0),
    ...     center_frequency=6.25e6,
    ...     sound_speed=1540.0,
    ...     sampling_frequency=25e6,
    ...     n_el=128,
    ...     n_tx=11,
    ...     probe_geometry=probe.probe_geometry,
    ... )

    >>> # Access a derived property (computed lazily)
    >>> grid = scan.grid  # shape: (grid_size_z, grid_size_x, 3)

    >>> # Select a subset of transmit events
    >>> _ = scan.set_transmits(3)  # Use 3 evenly spaced transmits
    >>> _ = scan.set_transmits([0, 2, 4])  # Use specific transmit indices
    >>> _ = scan.set_transmits("all")  # Use all transmits

"""

import numpy as np
from keras import ops

from zea import log
from zea.beamform.pfield import compute_pfield
from zea.beamform.pixelgrid import (
    cartesian_pixel_grid,
    check_for_aliasing,
    polar_pixel_grid,
)
from zea.display import (
    compute_scan_convert_2d_coordinates,
    compute_scan_convert_3d_coordinates,
)
from zea.internal.parameters import Parameters, cache_with_dependencies


class Scan(Parameters):
    """Represents an ultrasound scan configuration with computed properties.

    Args:
        grid_size_x (int): Grid width in pixels. For a cartesian grid, this is the lateral (x)
            pixels in the grid, set to prevent aliasing if not provided. For a polar grid, this can
            be thought of as the number for rays in the polar direction.
        grid_size_z (int): Grid height in pixels. This is the number of axial (z) pixels in the
            grid, set to prevent aliasing if not provided.
        sound_speed (float, optional): Speed of sound in the medium in m/s.
            Defaults to 1540.0.
        sampling_frequency (float): Sampling frequency in Hz.
        center_frequency (float): Transmit center frequency in Hz.
        demodulation_frequency (float, optional): Demodulation frequency in Hz.
        n_el (int): Number of elements in the transducer array.
        n_tx (int): Number of transmit events in the dataset.
        n_ax (int): Number of axial samples in the received signal.
        n_ch (int, optional): Number of channels (1 for RF, 2 for IQ data).
        xlims (tuple of float): Lateral (x) limits of the imaging region in
            meters (min, max).
        ylims (tuple of float, optional): Elevation (y) limits of the imaging
            region in meters (min, max).
        zlims (tuple of float): Axial (z) limits of the imaging region
            in meters (min, max).
        probe_geometry (np.ndarray): Element positions as array of shape (n_el, 3).
        polar_angles (np.ndarray): Polar angles for each transmit event in radians of shape (n_tx,).
            These angles are often used in 2D imaging.
        azimuth_angles (np.ndarray): Azimuth angles for each transmit event in radians
            of shape (n_tx,). These angles are often used in 3D imaging.
        t0_delays (np.ndarray): Transmit delays in seconds of
            shape (n_tx, n_el), shifted such that the smallest delay is 0.
        tx_apodizations (np.ndarray): Transmit apodizations of shape (n_tx, n_el).
        focus_distances (np.ndarray): Distance from the origin point on the transducer to where the
            beam comes to focus for each transmit in meters of shape (n_tx,).
        transmit_origins (np.ndarray): Transmit origins of shape (n_tx, 3).
        initial_times (np.ndarray): Initial times in seconds for each event of shape (n_tx,).
        bandwidth_percent (float, optional): Bandwidth as percentage of center
            frequency. Defaults to 200.0.
        time_to_next_transmit (np.ndarray): The time between subsequent
            transmit events of shape (n_frames, n_tx).
        tgc_gain_curve (np.ndarray): Time gain compensation (TGC) curve of shape (n_ax,).
        waveforms_one_way (np.ndarray): The one-way transmit waveforms of shape
            (n_waveforms, n_samples).
        waveforms_two_way (np.ndarray): The two-way transmit waveforms of shape
            (n_waveforms, n_samples).
        tx_waveform_indices (np.ndarray): Indices of the waveform used for each
            transmit event of shape (n_tx,).
        t_peak (np.ndarray, optional): The time of the peak of the pulse of every transmit waveform
            of shape (n_waveforms,).
        pixels_per_wavelength (int, optional): Number of pixels per wavelength.
            Defaults to 4.
        element_width (float, optional): Width of each transducer element in meters.
        resolution (float, optional): Resolution for scan conversion in mm / pixel.
            If None, it is calculated based on the input image.
        pfield_kwargs (dict, optional): Additional parameters for pressure field computation.
            See `zea.beamform.pfield.compute_pfield` for details.
        apply_lens_correction (bool, optional): Whether to apply lens correction to
            delays. Defaults to False.
        lens_thickness (float, optional): Thickness of the lens in meters.
        f_number (float, optional): F-number of the transducer. Defaults to 1.0.
        theta_range (tuple, optional): Range of theta angles for 3D imaging.
        phi_range (tuple, optional): Range of phi angles for 3D imaging.
        rho_range (tuple, optional): Range of rho (radial) distances for 3D imaging.
        fill_value (float, optional): Value to use for out-of-bounds pixels.
        attenuation_coef (float, optional): Attenuation coefficient in dB/(MHz*cm).
            Defaults to 0.0.
        selected_transmits (None, str, int, list, slice, or np.ndarray, optional):
            Specifies which transmit events to select.
            - None or "all": Use all transmits.
            - "center": Use only the center transmit.
            - int: Select this many evenly spaced transmits.
            - list/array: Use these specific transmit indices.
            - slice: Use transmits specified by the slice (e.g., slice(0, 10, 2)).
        grid_type (str, optional): Type of grid to use for beamforming.
            Can be "cartesian" or "polar". Defaults to "cartesian".
        dynamic_range (tuple, optional): Dynamic range for image display.
            Defined in dB as (min_dB, max_dB). Defaults to (-60, 0).
        distance_to_apex (float, optional): Distance from the transducer to the apex of the
            pixel grid. This property is used for polar grids. Will be computed automatically
            if not provided.
    """

    VALID_PARAMS = {
        # beamforming related parameters
        "grid_size_x": {"type": int},
        "grid_size_y": {"type": int},
        "grid_size_z": {"type": int},
        "xlims": {"type": (tuple, list)},
        "ylims": {"type": (tuple, list)},
        "zlims": {"type": (tuple, list)},
        "pixels_per_wavelength": {"type": int, "default": 4},
        "pfield_kwargs": {"type": dict, "default": {}},
        "apply_lens_correction": {"type": bool, "default": False},
        "lens_sound_speed": {"type": float},
        "lens_thickness": {"type": float},
        "grid_type": {"type": str, "default": "cartesian"},
        "polar_limits": {"type": (tuple, list)},
        "dynamic_range": {"type": (tuple, list)},
        "selected_transmits": {
            "type": (type(None), str, int, list, slice, np.ndarray),
            "default": None,
        },
        # acquisition parameters
        "sound_speed": {"type": float, "default": 1540.0},
        "sampling_frequency": {"type": float},
        "center_frequency": {"type": float},
        "n_frames": {"type": int},
        "n_el": {"type": int},
        "n_tx": {"type": int},
        "n_ax": {"type": int},
        "n_ch": {"type": int},
        "bandwidth_percent": {"type": float, "default": 200.0},
        "demodulation_frequency": {"type": float},
        "element_width": {"type": float},
        "attenuation_coef": {"type": float, "default": 0.0},
        "f_number": {"type": float, "default": 1.0},
        # array parameters
        "probe_geometry": {"type": np.ndarray},
        "polar_angles": {"type": np.ndarray},
        "azimuth_angles": {"type": np.ndarray},
        "t0_delays": {"type": np.ndarray},
        "tx_apodizations": {"type": np.ndarray},
        "focus_distances": {"type": np.ndarray},
        "transmit_origins": {"type": np.ndarray},
        "initial_times": {"type": np.ndarray},
        "time_to_next_transmit": {"type": np.ndarray},
        "tgc_gain_curve": {"type": np.ndarray},
        "waveforms_one_way": {"type": np.ndarray, "default": None},
        "waveforms_two_way": {"type": np.ndarray, "default": None},
        "tx_waveform_indices": {"type": np.ndarray},
        "t_peak": {"type": np.ndarray},
        # scan conversion parameters
        "theta_range": {"type": (tuple, list)},
        "phi_range": {"type": (tuple, list), "default": None},
        "rho_range": {"type": (tuple, list)},
        "fill_value": {"type": float},
        "resolution": {"type": float, "default": None},
        "distance_to_apex": {"type": float},
    }

    @cache_with_dependencies("probe_geometry")
    def aperture_size(self):
        """Calculate the aperture size (x,y,z) based on the probe geometry."""
        if "probe_geometry" in self._params:
            x_coords = self.probe_geometry[:, 0]
            y_coords = self.probe_geometry[:, 1]
            z_coords = self.probe_geometry[:, 2]
            aperture_width = x_coords.max() - x_coords.min()
            aperture_height = y_coords.max() - y_coords.min()
            aperture_depth = z_coords.max() - z_coords.min()
            return np.array([aperture_width, aperture_height, aperture_depth])
        return None

    @cache_with_dependencies("polar_limits", "aperture_size")
    def distance_to_apex(self):
        """Calculate the distance from the transducer to the apex of the pixel grid."""
        if "distance_to_apex" in self._params:
            return self._params["distance_to_apex"]
        if self.aperture_size is not None:
            max_angle = np.max(np.abs(self.polar_limits))
            t = np.tan(max_angle)
            if np.isclose(t, 0.0):
                return 0.0
            distance_to_apex = (self.aperture_size[0] / 2) / t
            return distance_to_apex
        return 0.0

    @cache_with_dependencies(
        "xlims",
        "ylims",
        "zlims",
        "grid_size_x",
        "grid_size_z",
        "grid_size_y",
        "grid_type",
        "is_3d",
        "polar_limits",
        "distance_to_apex",
    )
    def grid(self):
        """The beamforming grid of shape (grid_size_z, grid_size_x, 3)."""
        if self.grid_type == "polar":
            if self.is_3d:
                raise NotImplementedError("3D polar grids are not yet supported.")
            return polar_pixel_grid(
                self.polar_limits,
                self.zlims,
                self.grid_size_z,
                self.grid_size_x,
                self.distance_to_apex,
            )
        elif self.grid_type == "cartesian":
            return cartesian_pixel_grid(
                self.xlims,
                self.zlims,
                self.ylims,
                grid_size_z=self.grid_size_z,
                grid_size_x=self.grid_size_x,
                grid_size_y=self.grid_size_y,
            )
        else:
            raise ValueError(
                f"Unsupported grid type: {self.grid_type}. Supported types are "
                "'cartesian' and 'polar'."
            )

    @cache_with_dependencies("xlims", "wavelength", "pixels_per_wavelength")
    def grid_size_x(self):
        """Grid width in pixels. For a cartesian grid, this is the lateral (x) pixels in the grid,
        set to prevent aliasing if not provided. For a polar grid, this can be thought of as
        the number for rays in the polar direction.
        """
        grid_size_x = self._params.get("grid_size_x")
        if grid_size_x is not None:
            return grid_size_x

        width = self.xlims[1] - self.xlims[0]
        min_grid_size_x = int(np.ceil(width / (self.wavelength / self.pixels_per_wavelength)))
        return max(min_grid_size_x, 1)

    @cache_with_dependencies(
        "ylims",
        "wavelength",
        "pixels_per_wavelength",
    )
    def grid_size_y(self):
        """Grid height in pixels. For a cartesian grid, this is the vertical (y) pixels in the grid,
        set to prevent aliasing if not provided. For a polar grid, this can be thought of as
        the number for rays in the azimuthal direction.
        """
        grid_size_y = self._params.get("grid_size_y")
        if grid_size_y is not None:
            return grid_size_y

        height = self.ylims[1] - self.ylims[0]
        min_grid_size_y = int(np.ceil(height / (self.wavelength / self.pixels_per_wavelength)))
        return max(min_grid_size_y, 1)

    @cache_with_dependencies(
        "zlims",
        "wavelength",
        "pixels_per_wavelength",
    )
    def grid_size_z(self):
        """Grid depth in pixels. This is the number of axial (z) pixels in the grid,
        set to prevent aliasing if not provided."""
        grid_size_z = self._params.get("grid_size_z")
        if grid_size_z is not None:
            return grid_size_z

        depth = self.zlims[1] - self.zlims[0]
        min_grid_size_z = int(np.ceil(depth / (self.wavelength / self.pixels_per_wavelength)))
        return max(min_grid_size_z, 1)

    @cache_with_dependencies("sound_speed", "center_frequency")
    def wavelength(self):
        """Calculate the wavelength based on sound speed and transmit center frequency."""
        return self.sound_speed / self.center_frequency

    @cache_with_dependencies("zlims", "polar_limits", "probe_geometry")
    def xlims(self):
        """The x-limits of the beamforming grid [m]. If not explicitly set, it is computed based
        on the polar limits and probe geometry.
        """
        xlims = self._params.get("xlims")
        if xlims is None:
            radius = max(self.zlims)
            xlims_polar = (
                radius * np.cos(-np.pi / 2 + self.polar_limits[0]),
                radius * np.cos(-np.pi / 2 + self.polar_limits[1]),
            )
            xlims_plane = (min(self.probe_geometry[:, 0]), max(self.probe_geometry[:, 0]))
            xlims = (
                min(xlims_polar[0], xlims_plane[0]),
                max(xlims_polar[1], xlims_plane[1]),
            )
        return xlims

    @cache_with_dependencies("zlims", "grid_type", "azimuth_limits", "probe_geometry")
    def ylims(self):
        """The y-limits of the beamforming grid [m]. If not explicitly set, it is computed based
        on the azimuth limits and probe geometry.
        """
        ylims = self._params.get("ylims")
        if ylims is not None:
            return ylims

        # If ylims not set, compute based on azimuth limits and probe geometry
        radius = max(self.zlims)
        ylims_azimuth = (
            (0.0, 0.0)  # avoid numerical imprecision with np.cos(np.pi/2)
            if self.azimuth_limits[0] == self.azimuth_limits[1]
            else (
                radius * np.cos(-np.pi / 2 + self.azimuth_limits[0]),
                radius * np.cos(-np.pi / 2 + self.azimuth_limits[1]),
            )
        )
        ylims_plane = (min(self.probe_geometry[:, 1]), max(self.probe_geometry[:, 1]))
        ylims = (
            min(ylims_azimuth[0], ylims_plane[0]),
            max(ylims_azimuth[1], ylims_plane[1]),
        )
        return ylims

    @cache_with_dependencies("sound_speed", "sampling_frequency", "n_ax")
    def zlims(self):
        """The z-limits of the beamforming grid [m]."""
        zlims = self._params.get("zlims")
        if zlims is None:
            return [0, self.sound_speed * self.n_ax / self.sampling_frequency / 2]
        return zlims

    @cache_with_dependencies("grid", "grid_type", "distance_to_apex")
    def extent(self):
        """The extent of the beamforming grid in the format (xmin, xmax, zmax, zmin).
        Can be directly used with `plt.imshow(x, extent=scan.extent)` for visualization.
        """
        xlims = (self.grid[:, :, 0].min(), self.grid[:, :, 0].max())
        zlims = (self.grid[:, :, 2].min(), self.grid[:, :, 2].max())

        # For polar grids, adjust zlims to account for distance to apex
        if self.grid_type == "polar":
            zlims = (zlims[0] + self.distance_to_apex, zlims[1])

        return np.array([xlims[0], xlims[1], zlims[1], zlims[0]])

    @cache_with_dependencies("grid")
    def flatgrid(self):
        """The beamforming grid of shape (grid_size_z*grid_size_x*grid_size_y, 3)."""
        return self.grid.reshape(-1, 3)

    @cache_with_dependencies("grid_size_x", "grid_size_y", "grid_size_z")
    def is_3d(self):
        return self.grid_size_y > 1 and self.grid_size_x > 1 and self.grid_size_z > 1

    @property
    def n_tx_total(self):
        """The total number of transmits in the full dataset."""
        return self._params["n_tx"]

    @cache_with_dependencies("selected_transmits")
    def n_tx(self):
        """The number of currently selected transmits."""
        return len(self.selected_transmits)

    def set_transmits(self, selection):
        """Select which transmit events to use.

        This method provides flexible ways to select transmit events:

        Args:
            selection: Specifies which transmits to select:
                - None: Use all transmits
                - "all": Use all transmits
                - "center": Use only the center transmit
                - "focused": Use only focused transmits
                - "diverging": Use only diverging transmits
                - "plane": Use only plane wave transmits
                - int: Select this many evenly spaced transmits
                - list/array: Use these specific transmit indices
                - slice: Use transmits specified by the slice (e.g., slice(0, 10, 2))

        Returns:
            The current instance for method chaining.

        Raises:
            ValueError: If the selection is invalid or incompatible with the scan.
        """
        n_tx_total = self._params.get("n_tx")
        if n_tx_total is None:
            raise ValueError("n_tx must be set before calling set_transmits")

        # Handle array-like - convert to list of indices
        if isinstance(selection, np.ndarray):
            if len(selection.shape) == 0:
                # Handle scalar numpy array
                return self.set_transmits(int(selection))
            elif len(selection.shape) == 1:
                selection = selection.tolist()
            else:
                raise ValueError(f"Invalid array shape: {selection.shape}")

        # Handle None and "all" - use all transmits
        if selection is None or selection == "all":
            self._params["selected_transmits"] = list(range(n_tx_total))
            self._invalidate("selected_transmits")
            return self

        # Handle "center" - use center transmit
        if selection == "center":
            self._params["selected_transmits"] = [n_tx_total // 2]
            self._invalidate("selected_transmits")
            return self

        if selection == "focused":
            value = self._params.get("focus_distances")
            if value is None:
                raise ValueError("No focus distances provided, cannot select focused transmits")
            idx = np.where(value > 0)[0].tolist()
            if len(idx) == 0:
                raise ValueError("No focused transmits found.")
            self._params["selected_transmits"] = idx
            self._invalidate("selected_transmits")
            return self

        if selection == "diverging":
            value = self._params.get("focus_distances")
            if value is None:
                raise ValueError("No focus distances provided, cannot select diverging transmits")
            idx = np.where(value < 0)[0].tolist()
            if len(idx) == 0:
                raise ValueError("No diverging transmits found.")
            self._params["selected_transmits"] = idx
            self._invalidate("selected_transmits")
            return self

        if selection == "plane":
            value = self._params.get("focus_distances")
            if value is None:
                raise ValueError("No focus distances provided, cannot select plane wave transmits")
            idx = np.concatenate([np.where(value == 0)[0], np.where(np.isinf(value))[0]]).tolist()
            if len(idx) == 0:
                raise ValueError("No plane wave transmits found.")
            self._params["selected_transmits"] = idx
            self._invalidate("selected_transmits")
            return self

        # Handle integer - select evenly spaced transmits
        if isinstance(selection, (int, np.integer)):
            selection = int(selection)  # Convert numpy integer to Python int
            if selection <= 0:
                raise ValueError(f"Number of transmits must be positive, got {selection}")

            if selection > n_tx_total:
                raise ValueError(
                    f"Requested {selection} transmits exceeds available transmits ({n_tx_total})"
                )

            if selection == 1:
                self._params["selected_transmits"] = [n_tx_total // 2]
            else:
                # Compute evenly spaced indices
                tx_indices = np.linspace(0, n_tx_total - 1, selection)
                self._params["selected_transmits"] = list(np.rint(tx_indices).astype(int))

            self._invalidate("selected_transmits")
            return self

        # Handle slice - convert to list of indices
        if isinstance(selection, slice):
            selection = list(range(n_tx_total))[selection]

        # Handle list of indices
        if isinstance(selection, list):
            # Validate indices
            if not all(isinstance(i, (int, np.integer)) for i in selection):
                raise ValueError("All transmit indices must be integers")

            if any(i < 0 or i >= n_tx_total for i in selection):
                raise ValueError(f"Transmit indices must be between 0 and {n_tx_total - 1}")

            self._params["selected_transmits"] = [
                int(i) for i in selection
            ]  # Convert numpy integers to Python ints
            self._invalidate("selected_transmits")
            return self

        # Aliasing check
        check_for_aliasing(self)

        raise ValueError(f"Unsupported selection type: {type(selection)}")

    @cache_with_dependencies("center_frequency")
    def demodulation_frequency(self):
        """The demodulation frequency in Hz."""
        if self._params.get("demodulation_frequency") is not None:
            return self._params["demodulation_frequency"]

        return self.center_frequency

    @cache_with_dependencies("selected_transmits")
    def polar_angles(self):
        """Polar angles for each transmit event in radians of shape (n_tx,).
        These angles are often used in 2D imaging."""
        value = self._params.get("polar_angles")
        if value is None:
            return None

        return value[self.selected_transmits]

    @cache_with_dependencies("polar_angles")
    def polar_limits(self):
        """The limits of the polar angles, used for polar grids."""
        value = self._params.get("polar_limits")
        if value is None and self.polar_angles is not None:
            value = self.polar_angles.min(), self.polar_angles.max()
            diff = value[1] - value[0]
            # add 15% margin to the limits
            value = (value[0] - 0.15 * diff, value[1] + 0.15 * diff)
        return value

    @cache_with_dependencies("selected_transmits", "n_tx")
    def azimuth_angles(self):
        """Azimuth angles for each transmit event in radians
        of shape (n_tx,). These angles are often used in 3D imaging."""
        value = self._params.get("azimuth_angles")
        if value is None:
            log.warning("No azimuth angles provided, using zeros")
            return np.zeros(self.n_tx)

        return value[self.selected_transmits]

    @cache_with_dependencies("azimuth_angles")
    def azimuth_limits(self):
        """The limits of the azimuth angles."""
        value = self._params.get("azimuth_limits")
        if value is None and self.azimuth_angles is not None:
            value = self.azimuth_angles.min(), self.azimuth_angles.max()
            diff = value[1] - value[0]
            # add 15% margin to the limits
            value = (value[0] - 0.15 * diff, value[1] + 0.15 * diff)
        return value

    @cache_with_dependencies("selected_transmits", "n_el", "n_tx")
    def t0_delays(self):
        """Transmit delays in seconds of
        shape (n_tx, n_el), shifted such that the smallest delay is 0."""
        value = self._params.get("t0_delays")
        if value is None:
            log.warning("No transmit delays provided, using zeros")
            return np.zeros((self.n_tx, self.n_el))

        return value[self.selected_transmits]

    @cache_with_dependencies("selected_transmits", "n_el", "n_tx")
    def tx_apodizations(self):
        """Transmit apodizations of shape (n_tx, n_el)."""
        value = self._params.get("tx_apodizations")
        if value is None:
            log.warning("No transmit apodizations provided, using ones")
            return np.ones((self.n_tx, self.n_el))

        return value[self.selected_transmits]

    @cache_with_dependencies("selected_transmits", "n_tx")
    def focus_distances(self):
        """Focus distances in meters for each event of shape (n_tx,)."""
        value = self._params.get("focus_distances")
        if value is None:
            log.warning("No focus distances provided, using zeros")
            return np.zeros(self.n_tx)

        return value[self.selected_transmits]

    @cache_with_dependencies("selected_transmits", "n_tx")
    def transmit_origins(self):
        """Transmit origins of shape (n_tx, 3)."""
        value = self._params.get("transmit_origins")
        if value is None:
            log.warning("No transmit origins provided, using zeros")
            return np.zeros((self.n_tx, 3))

        return value[self.selected_transmits]

    @cache_with_dependencies("selected_transmits", "n_tx")
    def initial_times(self):
        """Initial times in seconds for each event of shape (n_tx,)."""
        value = self._params.get("initial_times")
        if value is None:
            log.warning("No initial times provided, using zeros")
            return np.zeros(self.n_tx)

        return value[self.selected_transmits]

    @cache_with_dependencies("waveforms_one_way", "waveforms_two_way")
    def n_waveforms(self):
        """The number of unique transmit waveforms."""

        if self.waveforms_one_way is not None:
            return self.waveforms_one_way.shape[0]

        if self.waveforms_two_way is not None:
            return self.waveforms_two_way.shape[0]

        return 1

    @cache_with_dependencies("center_frequency", "n_waveforms")
    def t_peak(self):
        """The time of the peak of the pulse in seconds of shape (n_waveforms,)."""
        t_peak = self._params.get("t_peak")
        if t_peak is None:
            t_peak = np.array([1 / self.center_frequency] * self.n_waveforms)

        return t_peak

    @cache_with_dependencies("selected_transmits")
    def time_to_next_transmit(self):
        """The time between subsequent transmit events of shape (n_frames, n_tx)."""
        value = self._params.get("time_to_next_transmit")
        if value is None:
            return None

        selected = self.selected_transmits
        return value[:, selected]

    @cache_with_dependencies("n_ax")
    def tgc_gain_curve(self):
        """Time gain compensation (TGC) curve of shape (n_ax,)."""
        value = self._params.get("tgc_gain_curve")
        if value is None:
            return np.ones(self.n_ax)
        return value[: self.n_ax]

    @cache_with_dependencies("selected_transmits", "n_tx")
    def tx_waveform_indices(self):
        """Indices of the waveform used for each transmit event of shape (n_tx,)."""
        value = self._params.get("tx_waveform_indices")
        if value is None:
            return np.zeros(self.n_tx, dtype=int)

        return value[self.selected_transmits]

    @cache_with_dependencies(
        "sound_speed",
        "center_frequency",
        "bandwidth_percent",
        "n_el",
        "probe_geometry",
        "tx_apodizations",
        "grid",
        "t0_delays",
        "pfield_kwargs",
    )
    def pfield(self) -> np.ndarray:
        """Compute or return the pressure field (pfield) for weighting."""
        assert not self.is_3d, "Pfield computation only supported for 2D scans"
        pfield = compute_pfield(
            sound_speed=self.sound_speed,
            center_frequency=self.center_frequency,
            bandwidth_percent=self.bandwidth_percent,
            n_el=self.n_el,
            probe_geometry=self.probe_geometry,
            tx_apodizations=self.tx_apodizations,
            grid=self.grid,
            t0_delays=self.t0_delays,
            **self.pfield_kwargs,
        )
        return ops.convert_to_numpy(pfield)

    @cache_with_dependencies("pfield")
    def flat_pfield(self):
        """Flattened pfield for weighting of shape (n_pix, n_tx)."""
        return self.pfield.reshape(self.n_tx, -1).swapaxes(0, 1)

    @cache_with_dependencies("zlims", "distance_to_apex")
    def rho_range(self):
        """A tuple specifying the range of rho values (min_rho, max_rho). Defined in mm.
        Used for scan conversion."""
        value = self._params.get("rho_range")
        if value is None:
            return (self.zlims[0], self.zlims[1] + self.distance_to_apex)
        return value

    @cache_with_dependencies("polar_limits")
    def theta_range(self):
        """A tuple specifying the range of theta values (min_theta, max_theta).
        Defined in radians. Used for scan conversion."""
        value = self._params.get("theta_range")
        if value is None and self.polar_limits is not None:
            return self.polar_limits
        return value

    @cache_with_dependencies(
        "rho_range", "theta_range", "resolution", "grid_size_z", "grid_size_x", "distance_to_apex"
    )
    def coordinates_2d(self):
        """The coordinates for scan conversion."""
        coords, _ = compute_scan_convert_2d_coordinates(
            (self.grid_size_z, self.grid_size_x),
            self.rho_range,
            self.theta_range,
            self.resolution,
            distance_to_apex=self.distance_to_apex,
        )
        return coords

    @cache_with_dependencies(
        "rho_range",
        "theta_range",
        "phi_range",
        "resolution",
        "grid_size_z",
        "grid_size_x",
    )
    def coordinates_3d(self):
        """The coordinates for scan conversion."""
        # TODO: no grid_size_y... this is broken
        coords, _ = compute_scan_convert_3d_coordinates(
            (self.grid_size_z, self.grid_size_x),
            self.rho_range,
            self.theta_range,
            self.phi_range,
            self.resolution,
        )
        return coords

    @cache_with_dependencies("phi_range", "coordinates_2d", "coordinates_3d")
    def coordinates(self):
        """Get the coordinates for scan conversion, will be 3D if phi_range is set,
        otherwise 2D."""
        return self.coordinates_3d if getattr(self, "phi_range", None) else self.coordinates_2d

    @cache_with_dependencies("time_to_next_transmit")
    def pulse_repetition_frequency(self):
        """The pulse repetition frequency (PRF) [Hz]. Assumes a constant PRF."""
        if self.time_to_next_transmit is None:
            log.warning("Time to next transmit is not set, cannot compute PRF")
            return None

        pulse_repetition_interval = np.mean(self.time_to_next_transmit)

        return 1 / pulse_repetition_interval

    @cache_with_dependencies("time_to_next_transmit")
    def frames_per_second(self):
        """The number of frames per second [Hz]. Assumes a constant frame rate.

        Frames per second computed based on time between transmits within a frame.
        Ignores time between frames (e.g. due to processing).

        Uses the time it took to do all transmits (per frame). So if you only use some portion
        of the transmits, the fps will still be calculated based on all.
        """
        time_to_next_transmit = self._params.get("time_to_next_transmit")
        if time_to_next_transmit is None:
            log.warning("Time to next transmit is not set, cannot compute fps")
            return None

        # Check if fps is constant
        uniq = np.unique(time_to_next_transmit, axis=0)  # frame axis
        if uniq.shape[0] != 1:
            log.warning("Time to next transmit is not constant")

        # Compute fps
        time = np.mean(np.sum(time_to_next_transmit, axis=1))
        fps = 1 / time
        return fps

    @cache_with_dependencies("probe_geometry")
    def element_width(self):
        """The width of each transducer element in meters."""
        value = self._params.get("element_width")
        if value is None:
            # assume uniform spacing
            return np.linalg.norm(self.probe_geometry[1] - self.probe_geometry[0])
        return value

    def __setattr__(self, key, value):
        if key == "selected_transmits":
            # If setting selected_transmits, call set_transmits to handle logic
            self.set_transmits(value)
        else:
            return super().__setattr__(key, value)



"""
Parameter management system for ultrasound imaging.

This module provides the :class:`Parameters` base class, which implements
dependency-tracked, type-checked, and cacheable parameter logic for scientific
applications, primarily to support :class:`zea.Scan`.

See the Parameters class docstring for details on features and usage.
"""

import functools
import inspect
from copy import deepcopy

import numpy as np

from zea import log
from zea.internal.core import Object as ZeaObject
from zea.internal.core import _to_tensor, hash_elements, serialize_elements


def cache_with_dependencies(*deps):
    """Decorator to mark a method as a computed property with dependencies."""

    def decorator(func):
        func._dependencies = deps

        @functools.wraps(func)
        def wrapper(self: Parameters):
            self._assert_dependencies_met(func.__name__)

            if func.__name__ in self._cache:
                # Check if dependencies changed for mutable parameters
                current_hash = self._current_dependency_hash(func.__name__)
                if current_hash == self._dependency_versions.get(func.__name__):
                    return self._cache[func.__name__]
                else:
                    self._invalidate(func.__name__)

            result = func(self)
            self._cache[func.__name__] = result
            self._dependency_versions[func.__name__] = self._current_dependency_hash(func.__name__)
            return result

        return property(wrapper)

    return decorator


class MissingDependencyError(ValueError):
    """Exception indicating that a dependency of an attribute was not met."""

    def __init__(self, attribute: str, missing_dependencies: set):
        super().__init__(
            f"Cannot access '{attribute}' due to missing dependencies: "
            + f"{sorted(missing_dependencies)}"
        )


class NoDependencyError(ValueError):
    """Exception indicating that an attribute has no dependencies defined."""

    def __init__(self, name: str):
        super().__init__(f"'{name}' is not a computed property with dependencies.")


class Parameters(ZeaObject):
    """Base class for parameters with dependencies.

    This class provides a robust parameter management system,
    supporting dependency tracking, lazy evaluation, and type validation.

    **Features:**

    - **Type Validation:** All parameters must be validated against their
      expected types as specified in the `VALID_PARAMS` dictionary.
      Setting a parameter to an invalid type raises a `TypeError`.

    - **Dependency Tracking:** Computed properties can declare dependencies on
      other parameters or properties using the `@cache_with_dependencies`
      decorator. The system automatically tracks and resolves these dependencies.

    - **Lazy Computation:** Computed properties are evaluated only when accessed,
      and their results are cached for efficiency.

    - **Cache Invalidation:** When a parameter changes, all dependent computed
      properties are invalidated and recomputed on next access.

    - **Leaf Parameter Enforcement:** Only leaf parameters
      (those directly listed in `VALID_PARAMS`) can be set. Attempting to set a computed
      property raises an informative `ValueError` listing the leaf parameters
      that must be changed instead.

    - **Optional Dependency Parameters:** Parameters can be both set directly (as a leaf)
      or computed from dependencies if not set. If a parameter is present in `VALID_PARAMS`
      and also decorated with `@cache_with_dependencies`, it will use the explicitly set
      value if provided, or fall back to the computed value if not set or set to `None`.
      If you set such a parameter after it has been computed, the explicitly set value
      will override the computed value and remain in effect until you set it back to `None`,
      at which point it will again be computed from its dependencies. This pattern is useful
      for parameters that are usually derived from other values, but can also be overridden
      directly when needed, and thus don't have a forced relationship with the dependencies.

    - **Tensor Conversion:** The `to_tensor` method converts all parameters and optionally all
      computed properties to tensors for machine learning workflows.

    - **Error Reporting:** If a computed property cannot be resolved due to missing dependencies,
      an informative `MissingDependencyError` is raised, listing the missing parameters.

    **Usage Example:**

    .. doctest::

        >>> class MyParams(Parameters):
        ...     VALID_PARAMS = {
        ...         "a": {"type": int, "default": 1},
        ...         "b": {"type": float, "default": 2.0},
        ...         "d": {"type": float},  # optional dependency
        ...     }

        ...     @cache_with_dependencies("a", "b")
        ...     def c(self):
        ...        return self.a + self.b

        ...     @cache_with_dependencies("a", "b")
        ...     def d(self):
        ...         if self._params.get("d") is not None:
        ...             return self._params["d"]
        ...         return self.a * self.b

        >>> p = MyParams(a=3)
        >>> print(p.c)  # Computes and caches c
        5.0
        >>> print(p.c)  # Returns cached value
        5.0

        # Changing a parameter invalidates the cache
        >>> p.a = 4
        >>> print(p.c)  # Recomputes c, now 4 + 2.0 = 6.0
        6.0

        >>> # You are not allowed to set computed properties
        >>> # p.c = 5  # Raises ValueError

        >>> # Now check out the optional dependency, this can be either
        >>> # set directly during initialization or computed from dependencies (default)
        >>> print(p.d)  # Returns 8 (=4 * 2.0)
        8.0
        >>> p = MyParams(a=3, d=9.99)
        >>> print(p.d)
        9.99

    """

    VALID_PARAMS = None

    def __init__(self, **kwargs):
        super().__init__()

        # Check if VALID_PARAMS is defined
        if self.VALID_PARAMS is None:
            raise NotImplementedError("VALID_PARAMS must be defined in subclasses of Parameters.")

        # Check if the definition of the class has circular dependencies
        for name in self.__class__.__dict__:
            self._check_for_circular_dependencies(name)

        # Check if all dependencies are valid parameters or computed properties
        # Will automatically catch typo's etc.
        self._check_validity_of_dependencies()

        # Internal state
        self._params = {}
        self._properties = self.get_properties()
        self._cache = {}
        self._dependency_versions = {}

        # Tensor cache stores converted tensors for parameters and computed properties
        # to avoid converting them multiple times if there are no changes.
        self._tensor_cache = {}

        # Initialize parameters with defaults
        for param, config in self.VALID_PARAMS.items():
            if param not in kwargs and "default" in config:
                # need to deepcopy in case default is mutable
                kwargs[param] = deepcopy(config["default"])

        # Set provided parameters
        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def _validate_parameter(cls, key, value):
        # Check if the parameter is valid
        if key not in cls.VALID_PARAMS:
            raise ValueError(
                f"Invalid parameter: {key}. Valid parameters are: {list(cls.VALID_PARAMS.keys())}"
            )

        # Cast the value if needed and possible
        expected_type = cls.VALID_PARAMS[key]["type"]
        if expected_type is not None and value is not None and not isinstance(value, expected_type):
            value = cls._cast(key, value)

        # Check again
        if expected_type is not None and value is not None and not isinstance(value, expected_type):
            allowed = cls._human_readable_type(expected_type)
            raise TypeError(
                f"Parameter '{key}' expected type {allowed}, got {type(value).__name__}"
            )

        return value

    @classmethod
    def _cast(cls, key, value):
        """Cast parameter to the expected type if 'cast_from' is specified.

        Additionally, int to float conversion is allowed implicitly."""
        # If the value is a single-element array, convert it to a scalar
        # If it's a numpy scalar, convert it to a native Python type
        if (isinstance(value, np.ndarray) and value.size == 1) or isinstance(value, np.generic):
            value = value.item()

        # Assume the key exists in VALID_PARAMS
        config = cls.VALID_PARAMS[key]

        if value is None:
            return value

        cast_to = config["type"]
        if isinstance(cast_to, tuple):
            raise ValueError(f"Casting to multiple types is not supported for parameter '{key}'.")

        if "cast_from" not in config:
            if isinstance(value, int) and cast_to is float:
                # Allow implicit conversion from int to float
                return float(value)
            return value

        cast_types = config["cast_from"]
        if not isinstance(cast_types, tuple):
            cast_types = (cast_types,)

        if any(isinstance(value, t) for t in cast_types):
            value = cast_to(value)

        return value

    @staticmethod
    def _human_readable_type(type):
        """Convert a type or tuple of types to a human-readable string."""
        return (
            type.__name__ if not isinstance(type, tuple) else ", ".join([t.__name__ for t in type])
        )

    def copy(self):
        """Return a deep copy of the Parameters object."""
        return self.__class__(**deepcopy(self._params))

    @property
    def serialized(self):
        """Compute the checksum of the object only if not already done"""
        if self._serialized is None:
            self._serialized = serialize_elements([self._params])
        return self._serialized

    @classmethod
    def _is_property_with_dependencies(cls, name):
        """Check if a class attribute is a property with dependencies."""
        attr = getattr(cls, name, None)
        return isinstance(attr, property) and hasattr(attr.fget, "_dependencies")

    @classmethod
    def _get_dependencies(cls, name):
        """Get the dependencies of a computed property."""
        if not cls._is_property_with_dependencies(name):
            raise NoDependencyError(name)
        return getattr(cls, name).fget._dependencies

    def _find_leaf_params(self, name, seen=None):
        """Recursively find all leaf parameters that a property depends on.

        If it is an optional dependency parameter, it will be included as a leaf. Not the ones it
        depends on.
        """
        if seen is None:
            seen = set()
        if name in seen:
            return set()
        seen.add(name)

        # If the name is already a leaf parameter, return it
        if name in self._params or name in self.VALID_PARAMS:
            return {name}

        # If the name is a property with dependencies, find its leaf parameters
        if self._is_property_with_dependencies(name):
            leaves = set()
            for dep in self._get_dependencies(name):
                leaves |= self._find_leaf_params(dep, seen)  # union
            return leaves

        raise AttributeError(f"'{name}' is not a valid parameter or computed property.")

    def _has_param(self, name):
        """Check if a parameter is set (i.e., exists in _params)."""
        # Check for existence of _params to avoid issues during unpickling
        return "_params" in self.__dict__ and name in self._params

    def __getattr__(self, item):
        """Handle attribute access for parameters only.

        Properties with dependencies are handled by cache_with_dependencies decorator.
        Regular properties are handled by normal Python descriptor protocol.
        """
        # Return parameter value if it exists
        if self._has_param(item):
            return self._params[item]

        # Attribute not found
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{item}'")

    def __setattr__(self, key, value):
        # Give clear error message on assignment to methods
        class_attr = getattr(self.__class__, key, None)
        if callable(class_attr):
            raise AttributeError(
                f"Cannot assign to method '{key}'. "
                f"'{key}' is a method, not an attribute. "
                f"To use it, call it as a function, e.g.: '{self.__class__.__name__}.{key}(...)'"
            )

        # Allow setting private attributes
        if key.startswith("_"):
            return super().__setattr__(key, value)

        # Give clear error message on assignment to computed properties
        if self._is_property_with_dependencies(key) and key not in self.VALID_PARAMS:
            leaf_params = sorted(self._find_leaf_params(key))
            raise ValueError(
                f"Cannot set computed property '{key}'. Only leaf parameters can be set. "
                f"To change '{key}', set one or more of its leaf parameters: {leaf_params}"
            )

        # Validate new value
        value = self._validate_parameter(key, value)

        # Set the parameter
        self._params[key] = value

        # Invalidate cache for this parameter if it is also a computed property
        self._invalidate(key)

    def __delattr__(self, name):
        # Allow deletion of parameters, but not properties
        if name in self._params:
            del self._params[name]
            self._invalidate(name)
        elif name in self.VALID_PARAMS:
            raise ValueError(f"Cannot delete parameter '{name}' because it is not set.")
        else:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    @classmethod
    def _check_for_circular_dependencies(cls, name, seen=None):
        """Check for circular dependencies in the dependency tree with a depth-first search."""
        if seen is None:
            seen = set()
        if name in seen:
            raise RuntimeError(f"Circular dependency detected for '{name}'")
        seen = seen.copy()
        seen.add(name)

        if cls._is_property_with_dependencies(name):
            for dep in cls._get_dependencies(name):
                cls._check_for_circular_dependencies(dep, seen)

    @classmethod
    def _check_validity_of_dependencies(cls):
        for dep in cls.get_properties_with_dependencies():
            for d in cls._get_dependencies(dep):
                if not (d in cls.VALID_PARAMS or cls._is_property_with_dependencies(d)):
                    raise ValueError(
                        f"Dependency '{d}' of computed property '{dep}' is not a valid "
                        f"parameter or computed property. This has to be fixed in the "
                        f"{cls.__name__} class definition."
                    )

    def _find_all_dependents(self, target, seen=None):
        """
        Find all computed properties that depend (directly or indirectly) on the target parameter
        with a global search. Returns a set of property names that depend on the target.
        """
        dependents = set()
        if seen is None:
            seen = set()
        if target in seen:
            return dependents
        seen.add(target)
        for name in self.__class__.__dict__:
            if self._is_property_with_dependencies(name):
                if target in self._get_dependencies(name):
                    dependents.add(name)
                    # Recursively add dependents of this property
                    dependents |= self._find_all_dependents(name, seen)
        return dependents

    def _invalidate(self, key):
        """Invalidate a specific cached computed property and its dependencies."""
        self._cache.pop(key, None)
        self._dependency_versions.pop(key, None)
        self._tensor_cache.pop(key, None)
        self._serialized = None  # see core object
        self._invalidate_dependents(key)

    def _invalidate_dependents(self, changed_key):
        """
        Invalidate all cached computed properties that (directly or indirectly)
        depend on the changed_key.
        """
        for key in self._find_all_dependents(changed_key):
            self._invalidate(key)

    def _current_dependency_hash(self, key) -> str:
        """Compute a hash representing the current state of the dependencies of key.

        Mainly needed to track changes in mutable parameters.
        """
        if not self._is_property_with_dependencies(key):
            raise NoDependencyError(key)
        deps = self._find_leaf_params(key)
        values = [self._params.get(dep) for dep in sorted(deps)]
        return hash_elements(values)

    def _assert_dependencies_met(self, name):
        """Assert that all dependencies for a computed property are met."""
        missing_set = self._find_missing_dependencies(name)
        if missing_set:
            raise MissingDependencyError(name, missing_set)

    def _find_missing_dependencies(self, name) -> set:
        missing_set = set()

        # Return immediately if already in params or cache
        if name in self._params or name in self._cache:
            return missing_set

        if self._is_property_with_dependencies(name):
            for dep in self._get_dependencies(name):
                missing_set |= self._find_missing_dependencies(dep)  # union
        else:
            missing_set.add(name)

        return missing_set

    @classmethod
    def get_properties(cls):
        """Get all properties of the class, including those with dependencies."""
        return [name for name, value in inspect.getmembers(cls) if isinstance(value, property)]

    @classmethod
    def get_properties_with_dependencies(cls):
        """Get all properties of the class that have dependencies."""
        return [name for name in cls.get_properties() if cls._is_property_with_dependencies(name)]

    def to_tensor(self, include=None, exclude=None, keep_as_is: list = None):
        """
        Convert parameters and computed properties to tensors.

        Only one of `include` or `exclude` can be set.

        Args:
            include ("all", or list): Only include these parameter/property names.
                If "all", include all available parameters (i.e. their dependencies are met).
                If specified, will take the intersection with possible parameters, so non-existing
                keys will be ignored. Default is "all".
            exclude (None or list): Exclude these parameter/property names.
                If provided, these keys will be excluded from the output.
            keep_as_is (list): List of parameter/property names that should not be converted to
                tensors, but included as-is in the output.
        """
        if include is None and exclude is None:
            include = "all"

        if include is not None and exclude is not None:
            raise ValueError("Only one of 'include' or 'exclude' can be set.")

        # Determine which keys to include
        param_keys = set(self._params.keys())
        property_keys = set(self._properties)
        all_keys = param_keys | property_keys

        if include == "all":
            keys = all_keys
        elif include is not None:
            # Filter include list to only existing keys
            keys = set(include).intersection(all_keys)
        elif exclude is not None:
            # Take all keys except those in exclude
            keys = all_keys - set(exclude)

        tensor_dict = {}
        # Convert parameters and computed properties to tensors
        for key in keys:
            # Get the value from params or computed properties
            # This is essential to trigger dependency checks
            try:
                val = getattr(self, key)
            except MissingDependencyError as exc:
                if include == "all" or exclude is not None:
                    # If we are including all, we can skip this key
                    continue
                else:
                    raise exc

            if key in self._tensor_cache:
                tensor_dict[key] = self._tensor_cache[key]
            else:
                tensor_val = _to_tensor(key, val, keep_as_is=keep_as_is)
                tensor_dict[key] = tensor_val
                self._tensor_cache[key] = tensor_val

        return tensor_dict

    def __repr__(self):
        param_lines = []
        for k, v in self._params.items():
            if v is None:
                continue

            # Handle arrays by showing their shape instead of content
            if isinstance(v, np.ndarray):
                param_lines.append(f"{k}=array(shape={v.shape})")
            else:
                param_lines.append(f"{k}={repr(v)}")

        param_str = ", ".join(param_lines)
        return f"{self.__class__.__name__}({param_str})"

    def __str__(self):
        param_lines = []
        for k, v in self._params.items():
            if v is None:
                continue

            # Handle arrays by showing their shape instead of content
            if isinstance(v, np.ndarray):
                param_lines.append(f"    {k}=array(shape={v.shape})")
            else:
                param_lines.append(f"    {k}={v}")

        param_str = ",\n".join(param_lines)
        return f"{self.__class__.__name__}(\n{param_str}\n)"

    @classmethod
    def standardize_params(cls, **kwargs) -> dict:
        """Return a dict with only valid parameters set and cast to the right type."""
        params = {}
        for parameter, value in kwargs.items():
            if parameter in cls.VALID_PARAMS:
                params[parameter] = value
            else:
                log.debug(f"Skipping invalid parameter '{parameter}'.")
        return params

    @classmethod
    def safe_initialize(cls, **kwargs):
        """Reduce kwargs to only valid parameters and convert types as needed."""
        params = cls.standardize_params(**kwargs)

        if len(params) == 0:
            log.info(f"Could not find proper scan parameters in {kwargs}.")
        return cls(**params)



"""Simple implementation of optimizers that support multi-backend autodiff."""

import keras


def adam(step_size, b1=0.9, b2=0.999, eps=1e-8):
    """Construct optimizer triple for Adam.

    Implementation adapted from `JAX's example <https://jax.readthedocs.io/en/latest/_modules/jax/example_libraries/optimizers.html#adam>`_
    See example usage: `JAX's example usage <https://jax.readthedocs.io/en/latest/jax.example_libraries.optimizers.html#jax.example_libraries.optimizers.adam>`_

    Args:
        step_size: positive scalar
        b1: optional, a positive scalar value for beta_1, the exponential decay rate
            for the first moment estimates (default 0.9).
        b2: optional, a positive scalar value for beta_2, the exponential decay rate
            for the second moment estimates (default 0.999).
        eps: optional, a positive scalar value for epsilon, a small constant for
            numerical stability (default 1e-8).

    Returns:
        An (init_fun, update_fun, get_params) triple.
    """

    def init(x0):
        m0 = keras.ops.zeros_like(x0)
        v0 = keras.ops.zeros_like(x0)
        i = 0
        return x0, m0, v0, i

    def update(g, state):
        """Update rule for Adam optimizer.

        Args:
            g (array): gradient
            state (tuple): state of the optimizer
                (x, m, v, i) = (parameter, first moment estimate,
                second moment estimate, iteration count)

        Returns:
            state: updated state
        """
        x, m, v, i = state
        m = (1 - b1) * g + b1 * m  # First  moment estimate.
        v = (1 - b2) * keras.ops.square(g) + b2 * v  # Second moment estimate.
        mhat = m / (1 - b1 ** (i + 1))  # Bias correction.
        vhat = v / (1 - b2 ** (i + 1))
        x = x - step_size * mhat / (keras.ops.sqrt(vhat) + eps)
        i = i + 1
        return x, m, v, i

    def get_params(state):
        """Returns just the parameter from the state."""
        x, *_ = state
        return x

    return init, update, get_params



import json
import os
import urllib.request
import zipfile
from pathlib import Path

import imageio
import numpy as np
from PIL import Image
from tqdm import tqdm

from zea import log

# Girder API base URL shared by CAMUS and CETUS collections
GIRDER_API = "https://humanheart-project.creatis.insa-lyon.fr/database/api/v1"


def sitk_load(filepath: str | Path, squeeze: bool = False):
    """Load a NIfTI/medical image using SimpleITK and return the array and metadata.

    Args:
        filepath: Path to the image file.
        squeeze: If True, squeeze singleton dimensions from the array.
            Defaults to False.

    Returns:
        Tuple of:
            - Image array. Shape depends on the input and the ``squeeze`` parameter.
            - Dictionary of metadata: ``origin``, ``spacing``, ``direction``, ``size``,
              ``dimension``, and a ``metadata`` sub-dict with all image metadata keys.
    """
    try:
        import SimpleITK as sitk
    except ImportError as exc:
        raise ImportError(
            "SimpleITK is not installed. "
            "Please install it with `pip install SimpleITK` to use this function."
        ) from exc

    image = sitk.ReadImage(str(filepath))

    all_metadata = {}
    for k in image.GetMetaDataKeys():
        all_metadata[k] = image.GetMetaData(k)

    metadata = {
        "origin": image.GetOrigin(),
        "spacing": image.GetSpacing(),
        "direction": image.GetDirection(),
        "size": image.GetSize(),
        "dimension": image.GetDimension(),
        "metadata": all_metadata,
    }

    im_array = sitk.GetArrayFromImage(image)
    if squeeze:
        im_array = np.squeeze(im_array)
    return im_array, metadata


def load_avi(file_path, mode="L"):
    """Load a .avi file and return a numpy array of frames.

    Args:
        filename (str): The path to the video file.
        mode (str, optional): Color mode: "L" (grayscale) or "RGB".
            Defaults to "L".

    Returns:
        numpy.ndarray: Array of frames (num_frames, H, W) or (num_frames, H, W, C)
    """
    frames = []
    with imageio.get_reader(file_path) as reader:
        for frame in reader:
            img = Image.fromarray(frame)
            img = img.convert(mode)
            img = np.array(img)
            frames.append(img)
    return np.stack(frames)


def unzip(src: str | Path, dataset: str) -> Path:
    """
    Checks if data folder exist in src.
    Otherwise, unzip dataset.zip in src.

    Args:
        src (str | Path): The source directory containing the zip file or unzipped folder.
        dataset (str): The name of the dataset to unzip.
            Options are "picmus", "camus", "echonet", "echonetlvh".

    Returns:
        Path: The path to the unzipped dataset directory.
    """
    src = Path(src)
    if dataset == "picmus":
        zip_name = "picmus.zip"
        folder_name = "archive_to_download"
        unzip_dir = src / folder_name
    elif dataset == "camus":
        zip_name = "CAMUS_public.zip"
        folder_name = "CAMUS_public"
        unzip_dir = src / folder_name
    elif dataset == "echonet":
        zip_name = "EchoNet-Dynamic.zip"
        folder_name = "EchoNet-Dynamic"
        unzip_dir = src / folder_name / "Videos"
    elif dataset == "echonetlvh":
        zip_name = "EchoNet-LVH.zip"
        folder_name = "Batch1"
        unzip_dir = src
    else:
        raise ValueError(f"Dataset {dataset} not recognized for unzip.")

    if (src / folder_name).exists():
        if dataset == "echonetlvh":
            # EchoNetLVH dataset unzips into four folders. Check they all exist.
            assert (src / "Batch2").exists(), f"Missing Batch2 folder in {src}."
            assert (src / "Batch3").exists(), f"Missing Batch3 folder in {src}."
            assert (src / "Batch4").exists(), f"Missing Batch4 folder in {src}."
            assert (src / "MeasurementsList.csv").exists(), (
                f"Missing MeasurementsList.csv in {src}."
            )
            log.info(f"Found Batch1, Batch2, Batch3, Batch4 and MeasurementsList.csv in {src}.")
        return unzip_dir

    zip_path = src / zip_name
    if not zip_path.exists():
        raise FileNotFoundError(f"Could not find {zip_name} or {folder_name} folder in {src}.")

    log.info(f"Unzipping {zip_path} to {src}...")
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(src)
    log.info("Unzipping completed.")
    log.info(f"Starting conversion from {src / folder_name}.")
    return unzip_dir


def download_from_girder(  # pragma: no cover
    collection_id: str,
    destination: str | Path,
    dataset_name: str,
    patients: list[int] | None = None,
    top_folder_name: str = "dataset",
) -> Path:
    """Download a dataset from the Girder server.

    Navigates the Girder collection to find patient folders and downloads
    all files for each patient. Existing files are skipped.

    Args:
        collection_id: Girder collection ID for the dataset.
        destination: Directory where the dataset will be downloaded.
        dataset_name: Human-readable name used in log messages
            (e.g. ``"CAMUS"`` or ``"CETUS"``).
        patients: Optional list of patient IDs to download.
            If None, all patients in the collection are downloaded.
        top_folder_name: Name of the top-level folder inside the collection
            that contains patient subfolders. Defaults to ``"dataset"``.

    Returns:
        Path to the downloaded dataset directory.
    """
    destination = Path(destination)
    destination.mkdir(parents=True, exist_ok=True)

    timeout = int(os.getenv("ZEA_DOWNLOAD_TIMEOUT", "60"))

    # Get top-level folders in the collection
    url = f"{GIRDER_API}/folder?parentType=collection&parentId={collection_id}&limit=50"
    with urllib.request.urlopen(url, timeout=timeout) as resp:
        folders = json.loads(resp.read())

    dataset_folder_id = None
    for folder in folders:
        if folder["name"] == top_folder_name:
            dataset_folder_id = folder["_id"]
            break

    if dataset_folder_id is None:
        raise RuntimeError(
            f"Could not find '{top_folder_name}' folder in {dataset_name} collection."
        )

    # Get patient folders (paginated — some datasets have >50 patients)
    patient_folders = []
    offset = 0
    page_size = 50
    while True:
        url = (
            f"{GIRDER_API}/folder?parentType=folder&parentId={dataset_folder_id}"
            f"&limit={page_size}&offset={offset}"
        )
        with urllib.request.urlopen(url, timeout=timeout) as resp:
            page = json.loads(resp.read())
        if not page:
            break
        patient_folders.extend(page)
        if len(page) < page_size:
            break
        offset += page_size

    if patients is not None:
        patient_set = set(patients)
        patient_folders = [
            pf for pf in patient_folders if int(pf["name"].removeprefix("patient")) in patient_set
        ]

    log.info(f"Downloading {len(patient_folders)} patients from {dataset_name} dataset...")

    for pf in tqdm(patient_folders, desc="Downloading patients"):
        patient_name = pf["name"]
        patient_dir = destination / patient_name
        patient_dir.mkdir(parents=True, exist_ok=True)

        # Get items (files) in the patient folder
        url = f"{GIRDER_API}/item?folderId={pf['_id']}&limit=50"
        with urllib.request.urlopen(url, timeout=timeout) as resp:
            items = json.loads(resp.read())

        for item in items:
            file_path = patient_dir / item["name"]
            if file_path.exists():
                log.debug(f"File {file_path} already exists when downloading. Skipping.")
                continue

            download_url = f"{GIRDER_API}/item/{item['_id']}/download"
            log.debug(f"Downloading {item['name']}...")
            with urllib.request.urlopen(download_url, timeout=timeout) as resp:
                file_path.write_bytes(resp.read())

    log.info(f"{dataset_name} dataset downloaded to {destination}")
    return destination
